from __future__ import annotations

import concurrent.futures
import functools
import json
import os
import queue
import threading
import time
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import urlencode, urlparse

import httpx
import websockets
import websockets.exceptions
from typing_extensions import Self
from websockets.sync.client import connect as websocket_connect

from . import api, lemur, types
from . import client as _client


class _TranscriptImpl:
    def __init__(
        self,
        *,
        client: _client.Client,
        transcript_id: Optional[str],
    ) -> None:
        self._client = client
        self.transcript_id = transcript_id

        self.transcript: Optional[types.TranscriptResponse] = None

    @property
    def config(self) -> types.TranscriptionConfig:
        "Returns the configuration from the internal Transcript object"

        return types.TranscriptionConfig(
            **self.transcript.dict(
                include=set(types.RawTranscriptionConfig.__fields__),
                exclude_none=True,
            )
        )

    @classmethod
    def from_response(
        cls,
        *,
        client: _client.Client,
        response: types.TranscriptResponse,
    ) -> Self:
        self = cls(
            client=client,
            transcript_id=response.id,
        )
        self.transcript = response

        return self

    def wait_for_completion(self) -> Self:
        """
        polls the given transcript until we have a status other than `processing` or `queued`
        """

        while True:
            # No try-except - if there is an HTTP error then surface it to user
            self.transcript = api.get_transcript(
                self._client.http_client,
                self.transcript_id,
            )

            if self.transcript.status in (
                types.TranscriptStatus.completed,
                types.TranscriptStatus.error,
            ):
                break

            time.sleep(self._client.settings.polling_interval)

        return self

    def export_subtitles_srt(
        self,
        *,
        chars_per_caption: Optional[int],
    ) -> str:
        return api.export_subtitles_srt(
            client=self._client.http_client,
            transcript_id=self.transcript.id,
            chars_per_caption=chars_per_caption,
        )

    def export_subtitles_vtt(
        self,
        *,
        chars_per_caption: Optional[int],
    ) -> str:
        return api.export_subtitles_vtt(
            client=self._client.http_client,
            transcript_id=self.transcript.id,
            chars_per_caption=chars_per_caption,
        )

    def word_search(
        self,
        *,
        words: List[str],
    ) -> List[types.WordSearchMatch]:
        response = api.word_search(
            client=self._client.http_client,
            transcript_id=self.transcript.id,
            words=words,
        )

        return response.matches

    def get_sentences(self) -> List[types.Sentence]:
        response = api.get_sentences(
            client=self._client.http_client,
            transcript_id=self.transcript.id,
        )

        return response.sentences

    def get_paragraphs(self) -> List[types.Paragraph]:
        response = api.get_paragraphs(
            client=self._client.http_client,
            transcript_id=self.transcript.id,
        )

        return response.paragraphs

    @functools.lru_cache
    def get_redacted_audio_url(self) -> str:
        """
        Retrieve the URL for the PII-redacted audio file, if `redact_pii_audio` was enabled on the `TranscriptionConfig`.
        Subsequent calls will return cached URL rather than requesting it from the API again.

        Returns: The URL of the redacted audio file.
        """
        if not self.config.redact_pii or not self.config.redact_pii_audio:
            raise ValueError(
                "Redacted audio is only available when `redact_pii` and `redact_pii_audio` are set to `True`."
            )

        while True:
            try:
                return api.get_redacted_audio(
                    client=self._client.http_client,
                    transcript_id=self.transcript_id,
                ).redacted_audio_url
            except types.RedactedAudioIncompleteError:
                time.sleep(self._client.settings.polling_interval)

    def save_redacted_audio(self, filepath: str):
        """
        Retrieve the PII-redacted audio file, if `redact_pii_audio` was enabled on the `TranscriptionConfig`

        Args:
            filepath: The path to save the redacted audio file to.
        """
        with httpx.stream(method="GET", url=self.get_redacted_audio_url()) as response:
            if response.status_code not in (httpx.codes.OK, httpx.codes.NOT_MODIFIED):
                raise types.RedactedAudioUnavailableError(
                    f"Fetching redacted audio failed with status code {response.status_code}"
                )
            with open(filepath, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)

    @classmethod
    def delete_by_id(cls, transcript_id: str) -> types.Transcript:
        client = _client.Client.get_default().http_client
        response = api.delete_transcript(client=client, transcript_id=transcript_id)

        return Transcript.from_response(client=client, response=response)


class Transcript(types.Sourcable):
    """
    Transcript object to perform operations on the actual transcript.
    """

    def __init__(
        self,
        transcript_id: Optional[str],
        client: Optional[_client.Client] = None,
    ) -> None:
        self._client = client or _client.Client.get_default()

        self._impl = _TranscriptImpl(
            client=self._client,
            transcript_id=transcript_id,
        )
        self._executor = concurrent.futures.ThreadPoolExecutor()

    def wait_for_completion(self) -> Self:
        self._impl.wait_for_completion()

        return self

    def wait_for_completion_async(
        self,
    ) -> concurrent.futures.Future[Self]:
        return self._executor.submit(self.wait_for_completion)

    @classmethod
    def from_response(
        cls,
        *,
        client: _client.Client,
        response: types.TranscriptResponse,
    ) -> Self:
        _impl = _TranscriptImpl.from_response(client=client, response=response)

        self = cls(
            client=client,
            transcript_id=response.id,
        )

        self._impl = _impl

        return self

    @classmethod
    def get_by_id(cls, transcript_id: str) -> Self:
        """Fetch an existing transcript. Blocks until the transcript is completed.

        Args:
            transcript_id: the id of the transcript to fetch

        Returns:
            The transcript object identified by the given id.
        """
        return cls(transcript_id=transcript_id).wait_for_completion()

    @classmethod
    def get_by_id_async(cls, transcript_id: str) -> concurrent.futures.Future[Self]:
        """Fetch an existing transcript asynchronously.

        Args:
            transcript_id: the id of the transcript to fetch

        Returns:
            A future that will resolve to the transcript object identified by the given id.
        """
        return cls(transcript_id=transcript_id).wait_for_completion_async()

    @classmethod
    def delete_by_id(cls, transcript_id: str) -> types.Transcript:
        """Delete an existing transcript. Blocks until the transcript is completed.

        Args:
            transcript_id: the id of the transcript to delete

        Returns:
            A transcript object identified by the given id, with relevant fields/attributes cleared.
        """
        return _TranscriptImpl.delete_by_id(transcript_id)

    @classmethod
    def delete_by_id_async(
        cls, transcript_id: str
    ) -> concurrent.futures.Future[types.Transcript]:
        """Delete an existing transcript asynchronously.

        Args:
            transcript_id: the id of the transcript to delete

        Returns:
            A future that will resolve to a transcript object identified by the given id, with relevant fields/attributes cleared.
        """

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            future_transcript = executor.submit(
                _TranscriptImpl.delete_by_id, transcript_id
            )
        return future_transcript

    @property
    def id(self) -> Optional[str]:
        "The unique identifier of your transcription"

        return self._impl.transcript_id

    @property
    def config(self) -> types.TranscriptionConfig:
        "Return the corresponding configurations for the given transcript."

        return self._impl.config

    @property
    def json_response(self) -> Optional[dict]:
        "The full JSON response associated with the transcript."

        return self._impl.transcript.dict()

    @property
    def audio_url(self) -> str:
        "The corresponding audio url"

        return self._impl.transcript.audio_url

    @property
    def speech_model(self) -> Optional[str]:
        "The speech model used for the transcription"
        return self._impl.transcript.speech_model

    @property
    def text(self) -> Optional[str]:
        "The text transcription of your media file"

        return self._impl.transcript.text

    @property
    def summary(self) -> Optional[str]:
        "The summarization of the transcript"

        return self._impl.transcript.summary

    @property
    def chapters(self) -> Optional[List[types.Chapter]]:
        "The list of auto-chapters results"

        return self._impl.transcript.chapters

    @property
    def content_safety(self) -> Optional[types.ContentSafetyResponse]:
        "The results from the content safety analysis"

        return self._impl.transcript.content_safety_labels

    @property
    def sentiment_analysis(self) -> Optional[List[types.Sentiment]]:
        "The list of sentiment analysis results"

        return self._impl.transcript.sentiment_analysis_results

    @property
    def entities(self) -> Optional[List[types.Entity]]:
        "The list of entity detection results"

        return self._impl.transcript.entities

    @property
    def iab_categories(self) -> Optional[types.IABResponse]:
        "The results from the IAB category detection"

        return self._impl.transcript.iab_categories_result

    @property
    def auto_highlights(self) -> Optional[types.AutohighlightResponse]:
        "The results from the auto-highlights model"

        return self._impl.transcript.auto_highlights_result

    @property
    def status(self) -> types.TranscriptStatus:
        "The current status of the transcript"

        return self._impl.transcript.status

    @property
    def error(self) -> Optional[str]:
        "The error message in case the transcription fails"

        return self._impl.transcript.error

    @property
    def words(self) -> Optional[List[types.Word]]:
        "The list of words in the transcript"

        return self._impl.transcript.words

    @property
    def utterances(self) -> Optional[List[types.Utterance]]:
        """
        When `dual_channel` or `speaker_labels` is enabled,
        a list of utterances in the transcript.
        """

        return self._impl.transcript.utterances

    @property
    def confidence(self) -> Optional[float]:
        "The confidence our model has in the transcribed text, between 0 and 1"

        return self._impl.transcript.confidence

    @property
    def audio_duration(self) -> Optional[int]:
        "The duration of the audio in seconds"

        return self._impl.transcript.audio_duration

    @property
    def webhook_status_code(self) -> Optional[int]:
        "The status code we received from your server when delivering your webhook"

        return self._impl.transcript.webhook_status_code

    @property
    def webhook_auth(self) -> Optional[bool]:
        "Whether the webhook was sent with an HTTP authentication header"

        return self._impl.transcript.webhook_auth

    @property
    def lemur(self) -> lemur.Lemur:
        """
        Access AssemblyAI's LeMUR features.
        """

        return lemur.Lemur(
            client=self._client,
            sources=[types.LemurSource(self)],
        )

    def export_subtitles_srt(
        self,
        chars_per_caption: Optional[int] = None,
    ) -> str:
        """
        You can export your complete transcripts in SRT format,
        to be plugged into a video player for subtitles and closed captions.

        Args:
            chars_per_caption: To control the maximum number of characters per caption

        Returns: A string containing the all subtitles in SRT format.
        """

        return self._impl.export_subtitles_srt(
            chars_per_caption=chars_per_caption,
        )

    def export_subtitles_vtt(
        self,
        chars_per_caption: Optional[int] = None,
    ) -> str:
        """
        You can export your complete transcripts in VTT format,
        to be plugged into a video player for subtitles and closed captions.

        Args:
            chars_per_caption: To control the maximum number of characters per caption

        Returns: A string containing the all subtitles in VTT format.
        """

        return self._impl.export_subtitles_vtt(
            chars_per_caption=chars_per_caption,
        )

    def word_search(
        self,
        words: List[str],
    ) -> List[types.WordSearchMatch]:
        """
        Once a transcript has been completed, you can search through the transcript for a specific set of keywords.
        You can search for individual words, numbers, or phrases containing up to five words or numbers.

        Args:
            words: A list of words, numbers, or phrases (containing up to five words or numbers)

        Returns: A list of matches
        """

        return self._impl.word_search(
            words=words,
        )

    def get_sentences(
        self,
    ) -> List[types.Sentence]:
        """
        Semantically segment your transcript into sentences to create more reader-friendly transcripts.

        Returns: A list of sentence objects.
        """

        return self._impl.get_sentences()

    def get_paragraphs(
        self,
    ) -> List[types.Paragraph]:
        """
        Semantically segment your transcript into paragraphs to create more reader-friendly transcripts.

        Returns: A list of paragraph objects.
        """

        return self._impl.get_paragraphs()

    def get_redacted_audio_url(self) -> str:
        """
        Retrieve the URL for the PII-redacted audio file, if `redact_pii_audio` was enabled on the `TranscriptionConfig`.
        Subsequent calls will return cached URL rather than requesting it from the API again.

        Returns: The URL of the redacted audio file.
        """
        return self._impl.get_redacted_audio_url()

    def save_redacted_audio(self, filepath: str):
        """
        Retrieve the PII-redacted audio file, if `redact_pii_audio` was enabled on the `TranscriptionConfig`

        Args:
            filepath: The path to save the redacted audio file to.
        """
        return self._impl.save_redacted_audio(filepath=filepath)


class _TranscriptGroupImpl:
    def __init__(
        self,
        *,
        transcript_ids: List[str],
        client: _client.Client,
    ) -> None:
        self._client = client
        self.transcripts: List[Transcript] = []

        for transcript_id in transcript_ids:
            self.add_transcript(transcript_id)

    @property
    def transcript_ids(self) -> List[str]:
        return [t.id for t in self.transcripts]

    def add_transcript(self, transcript: Union[Transcript, str]) -> None:
        if isinstance(transcript, Transcript):
            self.transcripts.append(transcript)
        elif isinstance(transcript, str):
            self.transcripts.append(
                Transcript(
                    client=self._client,
                    transcript_id=transcript,
                )
            )
        else:
            raise TypeError("Unsupported type for `transcript`")

        return self

    def wait_for_completion(self, return_failures) -> Union[None, List[str]]:
        transcripts: List[Transcript] = []
        failures: List[str] = []

        future_transcripts: Dict[concurrent.futures.Future[Transcript], str] = {}

        for transcript in self.transcripts:
            future = transcript.wait_for_completion_async()
            future_transcripts[future] = transcript

        finished_futures, _ = concurrent.futures.wait(future_transcripts)

        for future in finished_futures:
            try:
                transcripts.append(future.result())
            except types.TranscriptError as e:
                failures.append(str(e))

        self.transcripts = transcripts

        if return_failures:
            return failures


class TranscriptGroup:
    """
    A group of transcripts.

    Used when transcribing multiple transcripts at once.
    """

    def __init__(
        self,
        transcript_ids: List[str] = [],
        client: Optional[_client.Client] = None,
    ) -> None:
        self._client = client or _client.Client.get_default()

        self._impl = _TranscriptGroupImpl(
            transcript_ids=transcript_ids,
            client=self._client,
        )
        self._executor = concurrent.futures.ThreadPoolExecutor()

    @property
    def transcripts(self) -> List[Transcript]:
        """
        Returns the list of the transcripts within the `TranscriptGroup`
        """

        return self._impl.transcripts

    def __iter__(self) -> Iterator[Transcript]:
        """
        Iterate over the transcripts within the `TranscriptGroup`
        """

        return iter(self.transcripts)

    @classmethod
    def get_by_ids(cls, transcript_ids: List[str]) -> Self:
        return cls(transcript_ids=transcript_ids).wait_for_completion()

    @classmethod
    def get_by_ids_async(
        cls, transcript_ids: List[str]
    ) -> concurrent.futures.Future[Self]:
        return cls(transcript_ids=transcript_ids).wait_for_completion_async()

    @property
    def status(self) -> types.TranscriptStatus:
        """
        Return the status of the `TranscriptGroup`.

        e.g. if any of the transcripts is in `error` status, the whole `TranscriptGroup` will be in `error` status.
        """

        all_status = {t.status for t in self.transcripts}

        if any(s == types.TranscriptStatus.error for s in all_status):
            return types.TranscriptStatus.error
        elif any(s == types.TranscriptStatus.queued for s in all_status):
            return types.TranscriptStatus.queued
        elif any(s == types.TranscriptStatus.processing for s in all_status):
            return types.TranscriptStatus.processing
        elif all(s == types.TranscriptStatus.completed for s in all_status):
            return types.TranscriptStatus.completed

    @property
    def lemur(self) -> lemur.Lemur:
        """
        Access AssemblyAI's LeMUR functionality.
        """

        return lemur.Lemur(
            client=self._impl._client,
            sources=[types.LemurSource(t) for t in self.transcripts],
        )

    def add_transcript(
        self,
        transcript: Union[Transcript, str],
    ) -> Self:
        """
        Adds a transcript to the given `TranscriptGroup`

        Args:
            transcript: A `Transcript` object or the ID as a `str`
        """
        self._impl.add_transcript(transcript)

        return self

    def wait_for_completion(
        self,
        return_failures: Optional[bool] = False,
    ) -> Union[Self, Tuple[Self, List[str]]]:
        """
        Polls each transcript within the `TranscriptGroup`.

        Note - if an HTTP error is encountered when waiting for a Transcript in the TranscriptGroup, it will be popped from the group and added to the list of failures.
        You can return this list of failures with `return_failures=True`.

        Args:
            return_failures: Whether to return a list of errors for transcripts that failed due to HTTP errors.
        """
        if return_failures:
            failures = self._impl.wait_for_completion(return_failures=return_failures)
            return self, failures

        self._impl.wait_for_completion(return_failures=return_failures)

        return self

    def wait_for_completion_async(
        self,
        return_failures: Optional[bool] = False,
    ) -> Union[
        concurrent.futures.Future[Self],
        concurrent.futures.Future[Tuple[Self, List[str]]],
    ]:
        return self._executor.submit(
            self.wait_for_completion, return_failures=return_failures
        )


class _TranscriberImpl:
    """
    Implementation of the Transcriber class.
    """

    def __init__(
        self,
        *,
        client: _client.Client,
        config: types.TranscriptionConfig,
    ) -> None:
        self._client = client
        self.config = config

    def upload_file(self, data: Union[str, BinaryIO]) -> str:
        if isinstance(data, str):
            with open(data, "rb") as audio_file:
                return api.upload_file(
                    client=self._client.http_client,
                    audio_file=audio_file,
                )
        else:
            return api.upload_file(
                client=self._client.http_client,
                audio_file=data,
            )

    def transcribe_url(
        self,
        *,
        url: str,
        config: types.TranscriptionConfig,
        poll: bool,
    ) -> Transcript:
        transcript_request = types.TranscriptRequest(
            audio_url=url,
            **config.raw.dict(exclude_none=True),
        )
        # No try-except - if there is an HTTP error raise it to the user
        transcript = Transcript.from_response(
            client=self._client,
            response=api.create_transcript(
                client=self._client.http_client,
                request=transcript_request,
            ),
        )

        if poll:
            return transcript.wait_for_completion()

        return transcript

    def transcribe_file(
        self,
        *,
        data: Union[str, BinaryIO],
        config: types.TranscriptionConfig,
        poll: bool,
    ) -> Transcript:
        # Note: If uploading fails, it should raise an Exception to the user, hence no try-except here.
        audio_url = self.upload_file(data)

        return self.transcribe_url(
            url=audio_url,
            config=config,
            poll=poll,
        )

    def transcribe(
        self,
        data: Union[str, BinaryIO],
        config: Optional[types.TranscriptionConfig],
        poll: bool,
    ) -> Transcript:
        if config is None:
            config = self.config

        if isinstance(data, str) and urlparse(data).scheme in {"http", "https"}:
            return self.transcribe_url(
                url=data,
                config=config,
                poll=poll,
            )

        return self.transcribe_file(
            data=data,
            config=config,
            poll=poll,
        )

    def transcribe_group(
        self,
        *,
        data: List[Union[str, BinaryIO]],
        config: Optional[types.TranscriptionConfig],
        poll: bool,
        return_failures: Optional[bool] = False,
    ) -> Union[TranscriptGroup, Tuple[TranscriptGroup, List[str]]]:
        if config is None:
            config = self.config

        future_transcripts: Dict[concurrent.futures.Future[Transcript], str] = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            for d in data:
                transcript_future = executor.submit(
                    self.transcribe,
                    data=d,
                    config=config,
                    poll=False,
                )

                future_transcripts[transcript_future] = d

        finished_futures, _ = concurrent.futures.wait(future_transcripts)

        transcript_group = TranscriptGroup(
            client=self._client,
        )
        failures = []

        for future in finished_futures:
            try:
                transcript_group.add_transcript(future.result())
            except types.TranscriptError as e:
                failures.append(f"Error processing {future_transcripts[future]}: {e}")

        if poll and return_failures:
            transcript_group, completion_failures = (
                transcript_group.wait_for_completion(return_failures=return_failures)
            )
            failures.extend(completion_failures)
        elif poll:
            transcript_group = transcript_group.wait_for_completion(
                return_failures=return_failures
            )

        if return_failures:
            return transcript_group, failures
        else:
            return transcript_group

    def list_transcripts(
        self,
        params: Optional[types.ListTranscriptParameters],
    ) -> types.ListTranscriptResponse:
        return api.list_transcripts(client=self._client.http_client, params=params)


class Transcriber:
    """
    A transcriber used for transcribing URLs or local audio files.
    """

    def __init__(
        self,
        *,
        client: Optional[_client.Client] = None,
        config: Optional[types.TranscriptionConfig] = None,
        max_workers: Optional[int] = None,
    ) -> None:
        """
        Initializes the `Transcriber` with the given parameters.

        Args:
            `client`: The `Client` to use for the `Transcriber`. If `None` is given, the
                default settings for the `Client` will be used.
            `config`: The default configuration for the `Transcriber`. If `None` is given,
                the default configuration of a `TranscriptionConfig` will be used.
            `max_workers`: The maximum number of parallel jobs when using the `_async`
                methods on the `Transcriber`. By default it uses `os.cpu_count() - 1`

        Example:
            To use the `Transcriber` with the default settings, you can simply do:
            ```
            transcriber = aai.Transcriber()
            ```

            To use the `Transcriber` with a custom configuration, you can do:
            ```
            config = aai.TranscriptionConfig(punctuate=False, format_text=False)

            transcriber = aai.Transcriber(config=config)
            ```
        """
        self._client = client or _client.Client.get_default()

        self._impl = _TranscriberImpl(
            client=self._client,
            config=config or types.TranscriptionConfig(),
        )

        if not max_workers:
            max_workers = max(1, os.cpu_count() - 1)

        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers,
        )

    @property
    def config(self) -> types.TranscriptionConfig:
        """
        Returns the default configuration of the `Transcriber`.
        """
        return self._impl.config

    @config.setter
    def config(self, config: types.TranscriptionConfig) -> None:
        """
        Sets the default configuration of the `Transcriber`.

        Args:
            `config`: The new default configuration.
        """
        self._impl.config = config

    def upload_file(self, data: Union[str, BinaryIO]) -> str:
        """
        Uploads an audio file which can be specified as local path or binary object.

        Args:
            `data`: A local file (as path), or a binary object.

        Returns: The URL of the uploaded audio file.
        """
        return self._impl.upload_file(data=data)

    def upload_file_async(
        self, data: Union[str, BinaryIO]
    ) -> concurrent.futures.Future[str]:
        """
        Uploads an audio file which can be specified as local path or binary object.

        Args:
            `data`: A local file (as path), or a binary object.

        Returns: The URL of the uploaded audio file.
        """
        return self._executor.submit(
            self._impl.upload_file,
            data=data,
        )

    def submit(
        self,
        data: Union[str, BinaryIO],
        config: Optional[types.TranscriptionConfig] = None,
    ) -> Transcript:
        """
        Submits a transcription job without waiting for its completion.

        Args:
            data: An URL, a local file (as path), or a binary object.
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
        """
        return self._impl.transcribe(
            data=data,
            config=config,
            poll=False,
        )

    def submit_group(
        self,
        data: List[Union[str, BinaryIO]],
        config: Optional[types.TranscriptionConfig] = None,
        return_failures: Optional[bool] = False,
    ) -> Union[TranscriptGroup, Tuple[TranscriptGroup, List[str]]]:
        """
        Submits multiple transcription jobs without waiting for their completion.

        Args:
            data: A list of local paths, URLs, or binary objects (can be mixed).
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
            return_failures: Whether to include a list of errors for transcriptions that failed due to HTTP errors
        """
        return self._impl.transcribe_group(
            data=data,
            config=config,
            poll=False,
            return_failures=return_failures,
        )

    def transcribe(
        self,
        data: Union[str, BinaryIO],
        config: Optional[types.TranscriptionConfig] = None,
    ) -> Transcript:
        """
        Transcribes an audio file which can be specified as local path, URL, or binary object.

        Args:
            data: An URL, a local file (as path), or a binary object.
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
        """

        return self._impl.transcribe(
            data=data,
            config=config,
            poll=True,
        )

    def transcribe_async(
        self,
        data: Union[str, BinaryIO],
        config: Optional[types.TranscriptionConfig] = None,
    ) -> concurrent.futures.Future[Transcript]:
        """
        Transcribes an audio file which can be specified as local path, URL, or binary object.

        Args:
            data: An URL, a local file (as path), or a binary object.
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
        """

        return self._executor.submit(
            self._impl.transcribe,
            data=data,
            config=config,
            poll=True,
        )

    def transcribe_group(
        self,
        data: List[Union[str, BinaryIO]],
        config: Optional[types.TranscriptionConfig] = None,
        return_failures: Optional[bool] = False,
    ) -> Union[TranscriptGroup, Tuple[TranscriptGroup, List[str]]]:
        """
        Transcribes a list of files (as local paths, URLs, or binary objects).

        Args:
            data: A list of local paths, URLs, or binary objects (can be mixed).
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
            return_failures: Whether to include a list of errors for transcriptions that failed due to HTTP errors
        """

        return self._impl.transcribe_group(
            data=data,
            config=config,
            poll=True,
            return_failures=return_failures,
        )

    def transcribe_group_async(
        self,
        data: List[Union[str, BinaryIO]],
        config: Optional[types.TranscriptionConfig] = None,
        return_failures: Optional[bool] = False,
    ) -> Union[
        concurrent.futures.Future[TranscriptGroup],
        concurrent.futures.Future[Tuple[TranscriptGroup, List[str]]],
    ]:
        """
        Transcribes a list of files (as local paths, URLs, or binary objects) asynchronously.

        Args:
            data: A list of local paths, URLs, or binary objects (can be mixed).
            config: Transcription options and features. If `None` is given, the Transcriber's
                default configuration will be used.
            return_failures: Whether to include a list of errors for transcriptions that failed due to HTTP errors
        """

        return self._executor.submit(
            self._impl.transcribe_group,
            data=data,
            config=config,
            poll=True,
            return_failures=return_failures,
        )

    def list_transcripts(
        self,
        params: Optional[types.ListTranscriptParameters] = None,
    ) -> types.ListTranscriptResponse:
        """
        Retrieve a list of transcripts that were created. Transcripts are sorted from newest to oldest.

        Args:
            params: The parameters to filter the transcript list by.

        Returns: A page with a list of transcripts along with page details.

        To paginate over all pages, you can set the `ListTranscriptParameters.before_id`
        to the `before_id` of the `prev_url`. Example:
        ```
        transcriber = aai.Transcriber()
        params = aai.ListTranscriptParameters()
        page = transcriber.list_transcripts(params)
        while page.page_details.before_id_of_prev_url is not None:
            params.before_id = page.page_details.before_id_of_prev_url
            page = transcriber.list_transcripts(params)
        ```
        """
        return self._impl.list_transcripts(params=params)

    def list_transcripts_async(
        self,
        params: Optional[types.ListTranscriptParameters] = None,
    ) -> concurrent.futures.Future[types.ListTranscriptResponse]:
        """
        Retrieve a list of transcripts that were created. Transcripts are sorted from newest to oldest.

        Args:
            params: The parameters to filter the transcript list by.

        Returns: A page with a list of transcripts along with page details.
        """
        return self._executor.submit(self._impl.list_transcripts, params=params)


class _RealtimeTranscriberImpl:
    def __init__(
        self,
        *,
        on_data: Callable[[types.RealtimeTranscript], None],
        on_error: Callable[[types.RealtimeError], None],
        on_open: Optional[Callable[[types.RealtimeSessionOpened], None]],
        on_close: Optional[Callable[[], None]],
        sample_rate: int,
        word_boost: List[str],
        encoding: Optional[types.AudioEncoding] = None,
        token: Optional[str] = None,
        client: _client.Client,
        end_utterance_silence_threshold: Optional[int],
        disable_partial_transcripts: Optional[bool],
        on_extra_session_information: Optional[
            Callable[[types.RealtimeSessionInformation], None]
        ] = None,
    ) -> None:
        self._client = client
        self._websocket: Optional[websockets.sync.client.ClientConnection] = None

        self._on_open = on_open
        self._on_data = on_data
        self._on_error = on_error
        self._on_close = on_close
        self._sample_rate = sample_rate
        self._word_boost = word_boost
        self._encoding = encoding
        self._token = token
        self._end_utterance_silence_threshold = end_utterance_silence_threshold
        self._disable_partial_transcripts = disable_partial_transcripts
        self._on_extra_session_information = on_extra_session_information

        self._write_queue: queue.Queue[Union[bytes, Dict]] = queue.Queue()
        self._write_thread = threading.Thread(target=self._write)
        self._read_thread = threading.Thread(target=self._read)
        self._stop_event = threading.Event()

    def connect(
        self,
        timeout: Optional[float],
    ) -> None:
        """
        Connects to the real-time service.

        Args:
            `timeout`: The maximum time to wait for the connection to be established.
        """

        params: Dict[str, Any] = {
            "sample_rate": self._sample_rate,
        }
        if self._word_boost:
            params["word_boost"] = json.dumps(self._word_boost)
        if self._encoding:
            params["encoding"] = self._encoding.value
        if self._token:
            params["token"] = self._token
        if self._disable_partial_transcripts:
            params["disable_partial_transcripts"] = self._disable_partial_transcripts
        if self._on_extra_session_information:
            params["enable_extra_session_information"] = True

        websocket_base_url = self._client.settings.base_url.replace("https", "wss")

        additional_headers = None
        if self._token is None:
            additional_headers = {"Authorization": f"{self._client.settings.api_key}"}

        try:
            self._websocket = websocket_connect(
                f"{websocket_base_url}{api.ENDPOINT_REALTIME_WEBSOCKET}?{urlencode(params)}",
                additional_headers=additional_headers,
                open_timeout=timeout,
            )
        except Exception as exc:
            return self._on_error(
                types.RealtimeError(
                    f"Could not connect to the real-time service: {exc}"
                )
            )

        self._read_thread.start()
        self._write_thread.start()

        if self._end_utterance_silence_threshold is not None:
            self.configure_end_utterance_silence_threshold(
                self._end_utterance_silence_threshold
            )

    def stream(self, data: bytes) -> None:
        """
        Streams audio data to the real-time service by putting it into a queue.
        """

        self._write_queue.put(data)

    def configure_end_utterance_silence_threshold(
        self, threshold_milliseconds: int
    ) -> None:
        """
        Configures the end of utterance silence threshold.
        Can be called multiple times during a session at any point after the session starts.

        Args:
            `threshold_milliseconds`: The threshold in milliseconds.
        """

        self._write_queue.put(
            _RealtimeEndUtteranceSilenceThreshold(threshold_milliseconds).as_dict()
        )

    def force_end_utterance(self) -> None:
        """
        Forces the end of the current utterance.
        """

        self._write_queue.put(_RealtimeForceEndUtterance().as_dict())

    def close(self, terminate: bool = False) -> None:
        """
        Closes the connection to the real-time service gracefully.
        """
        if terminate and not self._stop_event.is_set():
            self._write_queue.put({"terminate_session": True})

        try:
            self._read_thread.join()
            self._write_thread.join()
            self._websocket.close()
        except Exception:
            pass

        if self._on_close:
            self._on_close()

    def _read(self) -> None:
        """
        Reads messages from the real-time service.

        Must run in a separate thread to avoid blocking the main thread.
        """

        while not self._stop_event.is_set():
            try:
                message = self._websocket.recv(timeout=1)
            except TimeoutError:
                continue
            except websockets.exceptions.ConnectionClosed as exc:
                return self._handle_error(exc)

            try:
                message = json.loads(message)
            except json.JSONDecodeError as exc:
                self._on_error(
                    types.RealtimeError(
                        f"Could not decode message: {exc}",
                    )
                )
                continue

            self._handle_message(message)

    def _write(self) -> None:
        """
        Writes messages to the real-time service.

        Must run in a separate thread to avoid blocking the main thread.
        """

        while not self._stop_event.is_set():
            try:
                data = self._write_queue.get(timeout=1)
            except queue.Empty:
                continue

            try:
                if isinstance(data, dict):
                    self._websocket.send(json.dumps(data))
                elif isinstance(data, bytes):
                    self._websocket.send(data)
                else:
                    raise ValueError("unsupported message type")
            except websockets.exceptions.ConnectionClosed as exc:
                return self._handle_error(exc)

    def _handle_message(
        self,
        message: Dict[str, Any],
    ) -> None:
        """
        Handles a message received from the real-time service by calling the appropriate
        callback.

        Args:
            `message`: The message to handle.
        """
        if "message_type" in message:
            if message["message_type"] == types.RealtimeMessageTypes.partial_transcript:
                self._on_data(types.RealtimePartialTranscript(**message))
            elif message["message_type"] == types.RealtimeMessageTypes.final_transcript:
                self._on_data(types.RealtimeFinalTranscript(**message))
            elif (
                message["message_type"] == types.RealtimeMessageTypes.session_begins
                and self._on_open
            ):
                self._on_open(types.RealtimeSessionOpened(**message))
            elif (
                message["message_type"] == types.RealtimeMessageTypes.session_terminated
            ):
                self._stop_event.set()
            elif (
                message["message_type"]
                == types.RealtimeMessageTypes.session_information
            ):
                self._on_extra_session_information(
                    types.RealtimeSessionInformation(**message)
                )
        elif "error" in message:
            self._on_error(types.RealtimeError(message["error"]))

    def _handle_error(self, error: websockets.exceptions.ConnectionClosed) -> None:
        """
        Handles a WebSocket error by calling the appropriate callback.

        See a list of errors here:

        - https://www.iana.org/assignments/websocket/websocket.xhtml#close-code-number
        - https://www.assemblyai.com/docs/Guides/real-time_streaming_transcription#closing-and-status-codes
        """
        if (
            error.code >= 4000
            and error.code <= 4999
            and error.code in types.RealtimeErrorMapping
        ):
            error_message = types.RealtimeErrorMapping[error.code]
        else:
            error_message = error.reason

        if error.code != 1000:
            self._on_error(types.RealtimeError(error_message))

        self.close()

    @classmethod
    def create_temporary_token(
        cls,
        expires_in: int,
        timeout: Optional[float] = None,
    ) -> str:
        """
        Request a temporary authentication token.

        Args:
            expires_in: The amount of time until the token expires in seconds.
            timeout: The timeout in seconds to wait for a response.
                A `timeout` of `None` means no timeout.

        Returns: The temporary authentication token.
        """

        return api.create_temporary_token(
            client=_client.Client.get_default().http_client,
            request=types.RealtimeCreateTemporaryTokenRequest(
                expires_in=expires_in,
            ),
            http_timeout=timeout,
        )


class _RealtimeForceEndUtterance:
    def as_dict(self) -> Dict[str, bool]:
        return {
            "force_end_utterance": True,
        }


class _RealtimeEndUtteranceSilenceThreshold:
    def __init__(self, threshold_milliseconds: int) -> None:
        self._value = threshold_milliseconds

    @property
    def value(self) -> int:
        return self._value

    def as_dict(self) -> Dict[str, int]:
        return {"end_utterance_silence_threshold": self._value}


class RealtimeTranscriber:
    def __init__(
        self,
        *,
        on_data: Callable[[types.RealtimeTranscript], None],
        on_error: Callable[[types.RealtimeError], None],
        on_open: Optional[Callable[[types.RealtimeSessionOpened], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
        sample_rate: int,
        word_boost: List[str] = [],
        encoding: Optional[types.AudioEncoding] = None,
        token: Optional[str] = None,
        client: Optional[_client.Client] = None,
        end_utterance_silence_threshold: Optional[int] = None,
        disable_partial_transcripts: Optional[bool] = None,
        on_extra_session_information: Optional[
            Callable[[types.RealtimeSessionInformation], None]
        ] = None,
    ) -> None:
        """
        Creates a new real-time transcriber.

        Args:
            `on_data`: The callback to call when a new transcript is received.
            `on_error`: The callback to call when an error occurs.
            `on_open`: (Optional) The callback to call when the connection to the real-time service opens.
            `on_close`: (Optional) The callback to call when the connection to the real-time service closes.
            `sample_rate`: The sample rate of the audio data.
            `word_boost`: (Optional) A list of words to boost transcription probability for.
            `encoding`: (Optional) The encoding of the audio data.
            `token`: (Optional) A temporary authentication token.
            `client`: (Optional) The client to use for the real-time service.
            `end_utterance_silence_threshold`: (Optional) The end utterance silence threshold in milliseconds.
            `disable_partial_transcripts`: (Optional) If set to `True`, only final transcripts will be received.
            `on_extra_session_information`: (Optional) The callback to call when a `SessionInformation` message is received.
                If this callback is set, the parameter `enable_extra_session_information` is sent to the API, and the client
                receives a `SessionInformation` message right before receiving the session termination message.
        """

        self._client = client or _client.Client.get_default(
            api_key_required=token is None
        )

        self._impl = _RealtimeTranscriberImpl(
            on_open=on_open,
            on_data=on_data,
            on_error=on_error,
            on_close=on_close,
            sample_rate=sample_rate,
            word_boost=word_boost,
            encoding=encoding,
            token=token,
            client=self._client,
            end_utterance_silence_threshold=end_utterance_silence_threshold,
            disable_partial_transcripts=disable_partial_transcripts,
            on_extra_session_information=on_extra_session_information,
        )

    def connect(
        self,
        timeout: Optional[float] = 10.0,
    ) -> None:
        """
        Connects to the real-time service.

        Args:
            `timeout`: The timeout in seconds to wait for the connection to be established.
                A `timeout` of `None` means no timeout.
        """

        self._impl.connect(timeout=timeout)

    def stream(
        self, data: Union[bytes, Generator[bytes, None, None], Iterable[bytes]]
    ) -> None:
        """
        Streams raw audio data to the real-time service.

        Args:
            `data`: Raw audio data in `bytes` or a generator/iterable of `bytes`.

        Note: Make sure that `data` matches the `sample_rate` that was given in the constructor.
        """
        if isinstance(data, bytes):
            self._impl.stream(data)
            return

        for chunk in data:
            self._impl.stream(chunk)

    def configure_end_utterance_silence_threshold(
        self, threshold_milliseconds: int
    ) -> None:
        """
        Configures the silence duration threshold used to detect the end of an utterance.
        In practice, it's used to tune how the transcriptions are split into final transcripts.
        Can be called multiple times during a session at any point after the session starts.

        Args:
            `threshold_milliseconds`: The threshold in milliseconds.
        """
        self._impl.configure_end_utterance_silence_threshold(threshold_milliseconds)

    def force_end_utterance(self) -> None:
        """
        Forces the end of the current utterance.
        After calling this method, the server will end the current utterance and return a final transcript.
        """
        self._impl.force_end_utterance()

    def close(self) -> None:
        """
        Closes the connection to the real-time service.
        """

        self._impl.close(terminate=True)

    @classmethod
    def create_temporary_token(
        cls,
        expires_in: int,
        timeout: Optional[float] = None,
    ) -> str:
        """
        Request a temporary authentication token.

        Example:
            To create a token, you can simply do:
            ```
            token = aai.RealtimeTranscriber.create_temporary_token(expires_in=360000)
            ```

        Args:
            expires_in: The amount of time until the token expires in seconds.
            timeout: The timeout in seconds to wait for a response.
                A `timeout` of `None` means no timeout.

        Returns: The temporary authentication token.
        """
        return _RealtimeTranscriberImpl.create_temporary_token(
            expires_in=expires_in, timeout=timeout
        )



import uuid

import httpx
import pytest
from pytest_httpx import HTTPXMock

import assemblyai as aai
from assemblyai.api import (
    ENDPOINT_LEMUR,
    ENDPOINT_LEMUR_BASE,
)
from tests.unit import factories

aai.settings.api_key = "test"


def test_lemur_single_question_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether asking a single question succeeds.
    """

    # create a mock response of a LemurQuestionResponse
    mock_lemur_answer = factories.generate_dict_factory(
        factories.LemurQuestionResponse
    )()

    # we only want to mock one answer
    mock_lemur_answer["response"] = [mock_lemur_answer["response"][0]]

    # prepare the question to be asked
    question = aai.LemurQuestion(
        question="Which cars do the callers want to buy?",
        context="Callers are interested in buying cars",
        answer_options=["Toyota", "Honda", "Ford", "Chevrolet"],
    )

    # update the mock question with the question
    mock_lemur_answer["response"][0]["question"] = question.question

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_answer,
    )

    transcript = aai.Transcript(str(uuid.uuid4()))

    # mimic the usage of the SDK
    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result = lemur.question(question)

    # check whether answer is not a list
    assert isinstance(result, aai.LemurQuestionResponse)

    answers = result.response

    # check the response
    assert answers[0].question == mock_lemur_answer["response"][0]["question"]
    assert answers[0].answer == mock_lemur_answer["response"][0]["answer"]

    assert result.usage == mock_lemur_answer["usage"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_single_question_succeeds_input_text(httpx_mock: HTTPXMock):
    """
    Tests whether asking a single question succeeds with input text.
    """

    # create a mock response of a LemurQuestionResponse
    mock_lemur_answer = factories.generate_dict_factory(
        factories.LemurQuestionResponse
    )()

    # we only want to mock one answer
    mock_lemur_answer["response"] = [mock_lemur_answer["response"][0]]

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_answer,
    )

    # prepare the question to be asked
    question = aai.LemurQuestion(
        question="Which cars do the callers want to buy?",
        context="Callers are interested in buying cars",
        answer_options=["Toyota", "Honda", "Ford", "Chevrolet"],
    )
    # test input_text input
    # mimic the usage of the SDK
    lemur = aai.Lemur()
    result = lemur.question(
        question, input_text="This transcript is a test transcript."
    )

    # check whether answer is not a list
    assert isinstance(result, aai.LemurQuestionResponse)

    answers = result.response

    # check the response
    assert answers[0].question == mock_lemur_answer["response"][0]["question"]
    assert answers[0].answer == mock_lemur_answer["response"][0]["answer"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_multiple_question_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether asking multiple questions succeeds.
    """

    # create a mock response of a LemurQuestionResponse
    mock_lemur_answer = factories.generate_dict_factory(
        factories.LemurQuestionResponse
    )()

    # prepare the questions to be asked
    questions = [
        aai.LemurQuestion(
            question="Which cars do the callers want to buy?",
        ),
        aai.LemurQuestion(
            question="What price range are the callers looking for?",
        ),
    ]

    # update the mock questions with the questions
    mock_lemur_answer["response"][0]["question"] = questions[0].question
    mock_lemur_answer["response"][1]["question"] = questions[1].question

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_answer,
    )

    transcript = aai.Transcript(str(uuid.uuid4()))

    # mimic the usage of the SDK
    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result = lemur.question(questions=questions)

    assert isinstance(result, aai.LemurQuestionResponse)

    answers = result.response
    # check whether answers is a list
    assert isinstance(answers, list)

    # check the response
    for idx, answer in enumerate(answers):
        assert answer.question == mock_lemur_answer["response"][idx]["question"]
        assert answer.answer == mock_lemur_answer["response"][idx]["answer"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_multiple_question_succeeds_input_text(httpx_mock: HTTPXMock):
    """
    Tests whether asking multiple questions succeeds.
    """

    # create a mock response of a LemurQuestionResponse
    mock_lemur_answer = factories.generate_dict_factory(
        factories.LemurQuestionResponse
    )()

    # prepare the questions to be asked
    questions = [
        aai.LemurQuestion(
            question="Which cars do the callers want to buy?",
        ),
        aai.LemurQuestion(
            question="What price range are the callers looking for?",
        ),
    ]

    # update the mock questions with the questions
    mock_lemur_answer["response"][0]["question"] = questions[0].question
    mock_lemur_answer["response"][1]["question"] = questions[1].question

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_answer,
    )

    # test input_text input
    # mimic the usage of the SDK
    lemur = aai.Lemur()
    result = lemur.question(
        questions, input_text="This transcript is a test transcript."
    )
    assert isinstance(result, aai.LemurQuestionResponse)

    answers = result.response
    # check whether answers is a list
    assert isinstance(answers, list)

    # check the response
    for idx, answer in enumerate(answers):
        assert answer.question == mock_lemur_answer["response"][idx]["question"]
        assert answer.answer == mock_lemur_answer["response"][idx]["answer"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_question_fails(httpx_mock: HTTPXMock):
    """
    Tests whether asking a question fails.
    """

    # prepare the question to be asked
    question = aai.LemurQuestion(
        question="Which cars do the callers want to buy?",
        context="Callers are interested in buying cars",
        answer_options=["Toyota", "Honda", "Ford", "Chevrolet"],
    )

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.INTERNAL_SERVER_ERROR,
        method="POST",
        json={"error": "something went wrong"},
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])

    with pytest.raises(aai.LemurError):
        lemur.question(question)

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_summarize_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether summarizing a transcript via LeMUR succeeds.
    """

    # create a mock response of a LemurSummaryResponse
    mock_lemur_summary = factories.generate_dict_factory(
        factories.LemurSummaryResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/summary",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_summary,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result = lemur.summarize(context="Callers asking for cars", answer_format="TLDR")

    assert isinstance(result, aai.LemurSummaryResponse)

    summary = result.response

    # check the response
    assert summary == mock_lemur_summary["response"]

    assert result.usage == mock_lemur_summary["usage"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_summarize_succeeds_input_text(httpx_mock: HTTPXMock):
    """
    Tests whether summarizing a transcript via LeMUR succeeds with input text.
    """

    # create a mock response of a LemurSummaryResponse
    mock_lemur_summary = factories.generate_dict_factory(
        factories.LemurSummaryResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/summary",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_summary,
    )

    # test input_text input
    lemur = aai.Lemur()
    result = lemur.summarize(
        context="Callers asking for cars", answer_format="TLDR", input_text="Test test"
    )

    assert isinstance(result, aai.LemurSummaryResponse)

    summary = result.response

    # check the response
    assert summary == mock_lemur_summary["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_summarize_fails(httpx_mock: HTTPXMock):
    """
    Tests whether summarizing a transcript via LeMUR fails.
    """

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/summary",
        status_code=httpx.codes.INTERNAL_SERVER_ERROR,
        method="POST",
        json={"error": "something went wrong"},
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])

    with pytest.raises(aai.LemurError):
        lemur.summarize(context="Callers asking for cars", answer_format="TLDR")

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_action_items_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether generating action items for a transcript via LeMUR succeeds.
    """

    # create a mock response of a LemurActionItemsResponse
    mock_lemur_action_items = factories.generate_dict_factory(
        factories.LemurActionItemsResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/action-items",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_action_items,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result = lemur.action_items(
        context="Customers asking for help with resolving their problem",
        answer_format="Three bullet points",
    )

    assert isinstance(result, aai.LemurActionItemsResponse)

    action_items = result.response

    # check the response
    assert action_items == mock_lemur_action_items["response"]

    assert result.usage == mock_lemur_action_items["usage"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_action_items_succeeds_input_text(httpx_mock: HTTPXMock):
    """
    Tests whether generating action items for a transcript via LeMUR succeeds.
    """

    # create a mock response of a LemurActionItemsResponse
    mock_lemur_action_items = factories.generate_dict_factory(
        factories.LemurActionItemsResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/action-items",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_action_items,
    )

    # test input_text input
    lemur = aai.Lemur()
    result = lemur.action_items(
        context="Customers asking for help with resolving their problem",
        answer_format="Three bullet points",
        input_text="Test test",
    )

    assert isinstance(result, aai.LemurActionItemsResponse)

    action_items = result.response

    # check the response
    assert action_items == mock_lemur_action_items["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_action_items_fails(httpx_mock: HTTPXMock):
    """
    Tests whether generating action items for a transcript via LeMUR fails.
    """

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/action-items",
        status_code=httpx.codes.INTERNAL_SERVER_ERROR,
        method="POST",
        json={"error": "something went wrong"},
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])

    with pytest.raises(aai.LemurError):
        lemur.action_items(
            context="Customers asking for help with resolving their problem",
            answer_format="Three bullet points",
        )

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_task_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether creating a task request succeeds.
    """

    # create a mock response of a LemurTaskResponse
    mock_lemur_task_response = factories.generate_dict_factory(
        factories.LemurTaskResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_task_response,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(
        sources=[aai.LemurSource(transcript)],
    )
    result = lemur.task(
        prompt="Create action items of the meeting", context="An important meeting"
    )

    # check the response
    assert isinstance(result, aai.LemurTaskResponse)

    assert result.response == mock_lemur_task_response["response"]

    assert result.usage == mock_lemur_task_response["usage"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_task_succeeds_input_text(httpx_mock: HTTPXMock):
    """
    Tests whether creating a task request succeeds.
    """

    # create a mock response of a LemurTaskResponse
    mock_lemur_task_response = factories.generate_dict_factory(
        factories.LemurTaskResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_task_response,
    )
    # test input_text input
    lemur = aai.Lemur()
    result = lemur.task(
        prompt="Create action items of the meeting", input_text="Test test"
    )

    # check the response
    assert isinstance(result, aai.LemurTaskResponse)

    assert result.response == mock_lemur_task_response["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


@pytest.mark.parametrize(
    "final_model",
    (
        aai.LemurModel.claude3_5_sonnet,
        aai.LemurModel.claude3_opus,
        aai.LemurModel.claude3_haiku,
        aai.LemurModel.claude3_sonnet,
        aai.LemurModel.claude2_1,
        aai.LemurModel.claude2_0,
        aai.LemurModel.default,
        aai.LemurModel.claude_instant1_2,
        aai.LemurModel.basic,
        aai.LemurModel.mistral7b,
    ),
)
def test_lemur_task_succeeds(final_model, httpx_mock: HTTPXMock):
    """
    Tests whether creating a task request succeeds with other models.
    """

    # create a mock response of a LemurTaskResponse
    mock_lemur_task_response = factories.generate_dict_factory(
        factories.LemurTaskResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_task_response,
    )
    # test input_text input
    lemur = aai.Lemur()
    result = lemur.task(
        final_model=final_model,
        prompt="Create action items of the meeting",
        context="An important meeting",
        input_text="Test test",
    )

    # check the response
    assert isinstance(result, aai.LemurTaskResponse)

    assert result.response == mock_lemur_task_response["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_ask_coach_fails(httpx_mock: HTTPXMock):
    """
    Tests whether creating a task request fails.
    """

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.INTERNAL_SERVER_ERROR,
        method="POST",
        json={"error": "something went wrong"},
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])

    with pytest.raises(aai.LemurError):
        lemur.task(prompt="Create action items of the meeting")

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_purge_request_data_succeeds(httpx_mock: HTTPXMock):
    """
    Tests whether LeMUR request purging succeeds.
    """

    # create a mock response of a LemurPurgeResponse
    mock_lemur_purge_response = factories.generate_dict_factory(
        factories.LemurPurgeResponse
    )()

    mock_request_id: str = str(uuid.uuid4())

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR_BASE}/{mock_request_id}",
        status_code=httpx.codes.OK,
        method="DELETE",
        json=mock_lemur_purge_response,
    )

    # mimic the usage of the SDK
    result = aai.Lemur.purge_request_data(request_id=mock_request_id)

    # check the response
    assert isinstance(result, aai.LemurPurgeResponse)

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_purge_request_data_fails(httpx_mock: HTTPXMock):
    """
    Tests whether LeMUR request purging fails.
    """

    # create a mock response of a LemurPurgeResponse
    mock_lemur_purge_response = factories.generate_dict_factory(
        factories.LemurPurgeResponse
    )()

    mock_request_id: str = str(uuid.uuid4())

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR_BASE}/{mock_request_id}",
        status_code=httpx.codes.INTERNAL_SERVER_ERROR,
        method="DELETE",
        json=mock_lemur_purge_response,
    )

    with pytest.raises(aai.LemurError):
        aai.Lemur.purge_request_data(mock_request_id)

    assert len(httpx_mock.get_requests()) == 1


def test_lemur_single_question_async_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether asking a single question succeeds when async is used.
    """

    # create a mock response of a LemurQuestionResponse
    mock_lemur_answer = factories.generate_dict_factory(
        factories.LemurQuestionResponse
    )()

    # we only want to mock one answer
    mock_lemur_answer["response"] = [mock_lemur_answer["response"][0]]

    # prepare the question to be asked
    question = aai.LemurQuestion(
        question="Which cars do the callers want to buy?",
        context="Callers are interested in buying cars",
        answer_options=["Toyota", "Honda", "Ford", "Chevrolet"],
    )

    # update the mock question with the question
    mock_lemur_answer["response"][0]["question"] = question.question

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/question-answer",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_answer,
    )

    transcript = aai.Transcript(str(uuid.uuid4()))

    # mimic the usage of the SDK
    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result_future = lemur.question_async(question)

    result = result_future.result()

    # check whether answer is not a list
    assert isinstance(result, aai.LemurQuestionResponse)

    answers = result.response

    # check the response
    assert answers[0].question == mock_lemur_answer["response"][0]["question"]
    assert answers[0].answer == mock_lemur_answer["response"][0]["answer"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_summarize_async_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether summarizing a transcript via LeMUR succeeds
    when async is used.
    """

    # create a mock response of a LemurSummaryResponse
    mock_lemur_summary = factories.generate_dict_factory(
        factories.LemurSummaryResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/summary",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_summary,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result_future = lemur.summarize_async(
        context="Callers asking for cars", answer_format="TLDR"
    )

    result = result_future.result()

    assert isinstance(result, aai.LemurSummaryResponse)

    summary = result.response

    # check the response
    assert summary == mock_lemur_summary["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_action_items_async_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether generating action items for a transcript via LeMUR succeeds
    when async is used.
    """

    # create a mock response of a LemurActionItemsResponse
    mock_lemur_action_items = factories.generate_dict_factory(
        factories.LemurActionItemsResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/action-items",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_action_items,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(sources=[aai.LemurSource(transcript)])
    result_future = lemur.action_items_async(
        context="Customers asking for help with resolving their problem",
        answer_format="Three bullet points",
    )

    result = result_future.result()

    assert isinstance(result, aai.LemurActionItemsResponse)

    action_items = result.response

    # check the response
    assert action_items == mock_lemur_action_items["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_task_async_succeeds_transcript(httpx_mock: HTTPXMock):
    """
    Tests whether creating a task request succeeds when async is used.
    """

    # create a mock response of a LemurTaskResponse
    mock_lemur_task_response = factories.generate_dict_factory(
        factories.LemurTaskResponse
    )()

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_task_response,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(
        sources=[aai.LemurSource(transcript)],
    )
    result_future = lemur.task_async(prompt="Create action items of the meeting")

    result = result_future.result()

    # check the response
    assert isinstance(result, aai.LemurTaskResponse)

    assert result.response == mock_lemur_task_response["response"]

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_purge_request_data_async_succeeds(httpx_mock: HTTPXMock):
    """
    Tests whether LeMUR request purging succeeds when async is used...
    """

    # create a mock response of a LemurPurgeResponse
    mock_lemur_purge_response = factories.generate_dict_factory(
        factories.LemurPurgeResponse
    )()

    mock_request_id: str = str(uuid.uuid4())

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR_BASE}/{mock_request_id}",
        status_code=httpx.codes.OK,
        method="DELETE",
        json=mock_lemur_purge_response,
    )

    # mimic the usage of the SDK
    result_future = aai.Lemur.purge_request_data_async(request_id=mock_request_id)

    result = result_future.result()

    # check the response
    assert isinstance(result, aai.LemurPurgeResponse)

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


def test_lemur_usage_data(httpx_mock: HTTPXMock):
    """
    Tests whether usage data is correctly returned.
    """

    # create a mock response of a LemurTaskResponse
    mock_lemur_task_response = factories.generate_dict_factory(
        factories.LemurTaskResponse
    )()
    mock_lemur_task_response["usage"]["input_tokens"] = 100
    mock_lemur_task_response["usage"]["output_tokens"] = 200

    # mock the specific endpoints
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR}/task",
        status_code=httpx.codes.OK,
        method="POST",
        json=mock_lemur_task_response,
    )

    # mimic the usage of the SDK
    transcript = aai.Transcript(str(uuid.uuid4()))

    lemur = aai.Lemur(
        sources=[aai.LemurSource(transcript)],
    )
    result = lemur.task(prompt="Create action items of the meeting")

    # check the response
    assert isinstance(result, aai.LemurTaskResponse)

    assert result.usage == mock_lemur_task_response["usage"]
    assert (
        result.usage.input_tokens == mock_lemur_task_response["usage"]["input_tokens"]
    )
    assert (
        result.usage.output_tokens == mock_lemur_task_response["usage"]["output_tokens"]
    )

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1


@pytest.mark.parametrize("response_type", ("string_response", "qa_response"))
def test_lemur_get_response_data(response_type, httpx_mock: HTTPXMock):
    """
    Tests whether a LeMUR response data is correctly returned.
    """
    request_id = "1234"

    # create a mock response
    if response_type == "string_response":
        mock_lemur_response = factories.generate_dict_factory(
            factories.LemurStringResponse
        )()
    else:
        mock_lemur_response = factories.generate_dict_factory(
            factories.LemurQuestionResponse
        )()

    mock_lemur_response["request_id"] = request_id

    # mock the specific endpoint
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_LEMUR_BASE}/{request_id}",
        status_code=httpx.codes.OK,
        method="GET",
        json=mock_lemur_response,
    )

    # mimic the usage of the SDK
    lemur = aai.Lemur()
    result = lemur.get_response_data(request_id)

    # check the response
    if response_type == "string_response":
        assert isinstance(result, aai.LemurStringResponse)
    else:
        assert isinstance(result, aai.LemurQuestionResponse)

    assert result.request_id == request_id

    # check whether we mocked everything
    assert len(httpx_mock.get_requests()) == 1



import httpx
import pytest
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

import tests.unit.unit_test_utils as unit_test_utils
import assemblyai as aai
from assemblyai.api import ENDPOINT_TRANSCRIPT
from tests.unit import factories

aai.settings.api_key = "test"


class TranscriptWithPIIRedactionResponseFactory(
    factories.TranscriptCompletedResponseFactory
):
    redact_pii = True
    redact_pii_audio = True
    redact_pii_policies = [
        aai.types.PIIRedactionPolicy.date,
    ]


def test_redact_pii_disabled_by_default(httpx_mock: HTTPXMock):
    """
    Tests that excluding `redact_pii` from the `TranscriptionConfig` will
    result in the default behavior of it being excluded from the request body
    """
    request_body, _ = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),
        config=aai.TranscriptionConfig(),
    )
    assert request_body.get("redact_pii") is None
    assert request_body.get("redact_pii_audio") is None
    assert request_body.get("redact_pii_policies") is None
    assert request_body.get("redact_pii_sub") is None


def test_redact_pii_enabled(httpx_mock: HTTPXMock):
    """
    Tests that enabling `redact_pii`, along with the required `redact_pii_policies`
    parameter will result in the request body containing those fields
    """
    policies = [
        aai.types.PIIRedactionPolicy.date,
        aai.types.PIIRedactionPolicy.phone_number,
    ]

    request_body, _ = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=policies,
        ),
    )

    assert request_body.get("redact_pii") is True
    assert request_body.get("redact_pii_policies") == policies


def test_redact_pii_enabled_with_optional_params(httpx_mock: HTTPXMock):
    """
    Tests that enabling `redact_pii`, along with the other optional parameters
    relevant to PII redaction, will result in the request body containing
    those fields
    """
    policies = [
        aai.types.PIIRedactionPolicy.date,
        aai.types.PIIRedactionPolicy.phone_number,
    ]
    sub_type = aai.types.PIISubstitutionPolicy.entity_name

    request_body, _ = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_audio=True,
            redact_pii_policies=policies,
            redact_pii_sub=sub_type,
        ),
    )

    assert request_body.get("redact_pii") is True
    assert request_body.get("redact_pii_audio") is True
    assert request_body.get("redact_pii_policies") == policies
    assert request_body.get("redact_pii_sub") == sub_type


def test_redact_pii_fails_without_policies(httpx_mock: HTTPXMock):
    """
    Tests that enabling `redact_pii` without specifying any policies
    will result in an exception being raised before the API call is made
    """
    with pytest.raises(ValueError) as error:
        unit_test_utils.submit_mock_transcription_request(
            httpx_mock,
            mock_response={},
            config=aai.TranscriptionConfig(
                redact_pii=True,
                # No policies!
            ),
        )

    assert "policy" in str(error)

    # Check that the error was raised before any requests were made
    assert len(httpx_mock.get_requests()) == 0


def test_redact_pii_params_excluded_when_disabled(httpx_mock: HTTPXMock):
    """
    Tests that additional PII redaction parameters are excluded from the submission
    request body if `redact_pii` itself is not enabled.
    """
    request_body, _ = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=False,
            redact_pii_audio=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_sub=aai.types.PIISubstitutionPolicy.entity_name,
        ),
    )

    assert request_body.get("redact_pii") is None
    assert request_body.get("redact_pii_audio") is None
    assert request_body.get("redact_pii_policies") is None
    assert request_body.get("redact_pii_sub") is None


def __get_redacted_audio_api_url(transcript: aai.Transcript) -> str:
    return (
        f"{aai.settings.base_url}{ENDPOINT_TRANSCRIPT}/{transcript.id}/redacted-audio"
    )


REDACTED_AUDIO_URL = "https://example.org/redacted-audio.wav"


def __mock_successful_pii_audio_responses(
    httpx_mock: HTTPXMock, transcript: aai.Transcript
):
    # Mock pending redacted audio response on first call
    httpx_mock.add_response(
        url=__get_redacted_audio_api_url(transcript),
        status_code=202,
        method="GET",
        json={},
    )

    # Mock completed redacted audio response on second call
    httpx_mock.add_response(
        url=__get_redacted_audio_api_url(transcript),
        status_code=httpx.codes.OK,
        method="GET",
        json={
            "redacted_audio_url": REDACTED_AUDIO_URL,
            "status": "redacted_audio_ready",
        },
    )


def __mock_failed_pii_audio_responses(
    httpx_mock: HTTPXMock, transcript: aai.Transcript
):
    httpx_mock.add_response(
        url=__get_redacted_audio_api_url(transcript),
        status_code=400,
        method="GET",
        json={},
    )


def test_get_pii_redacted_audio_url(httpx_mock: HTTPXMock):
    """
    Tests that the PII-redacted audio URL can be retrieved from the API
    with a successful response
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_audio=True,
        ),
    )

    __mock_successful_pii_audio_responses(httpx_mock, transcript)
    redacted_audio_url = transcript.get_redacted_audio_url()

    # Ensure we made a third and fourth network request to get the redacted audio information
    assert len(httpx_mock.get_requests()) == 4

    assert redacted_audio_url == REDACTED_AUDIO_URL


def test_get_pii_redacted_audio_url_fails_if_redact_pii_not_enabled_for_transcript(
    httpx_mock: HTTPXMock,
):
    """
    Tests that an error is thrown before any requests are made if
    `redact_pii` was not enabled for the transcript and
    `get_redacted_audio_url` is called
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),  # standard response
        config=aai.TranscriptionConfig(),  # blank config
    )

    with pytest.raises(ValueError) as error:
        transcript.get_redacted_audio_url()

    assert "redact_pii" in str(error)

    # Ensure we never made the additional requests to get the redacted audio information
    assert len(httpx_mock.get_requests()) == 2


def test_get_pii_redacted_audio_url_fails_if_redact_pii_audio_not_enabled_for_transcript(
    httpx_mock: HTTPXMock,
):
    """
    Tests that an error is thrown before any requests are made if
    `redact_pii_audio` was not enabled for the transcript and
    `get_redacted_audio_url` is called
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response={
            **factories.generate_dict_factory(
                TranscriptWithPIIRedactionResponseFactory
            )(),
            "redact_pii_audio": False,
        },
        config=aai.TranscriptionConfig(
            redact_pii=True, redact_pii_policies=[aai.types.PIIRedactionPolicy.date]
        ),
    )

    with pytest.raises(ValueError) as error:
        transcript.get_redacted_audio_url()

    assert "redact_pii_audio" in str(error)

    # Ensure we never made the additional requests to get the redacted audio information
    assert len(httpx_mock.get_requests()) == 2


def test_get_pii_redacted_audio_url_fails_if_bad_response(httpx_mock: HTTPXMock):
    """
    Tests that `get_redacted_audio_url` raises a `RedactedAudioUnavailableError` if
    the request to fetch the redacted audio URL returns a `400` status code, indicating
    that the redacted audio has expired
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_audio=True,
        ),
    )

    __mock_failed_pii_audio_responses(httpx_mock, transcript)
    with pytest.raises(aai.types.RedactedAudioExpiredError):
        transcript.get_redacted_audio_url()


def test_save_pii_redacted_audio(httpx_mock: HTTPXMock, mocker: MockerFixture):
    """
    Tests that calling `save_redacted_audio` will download the redacted audio file
    to the caller's file system
    """

    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_audio=True,
        ),
    )

    # Mock response that returns the redacted-audio URL
    __mock_successful_pii_audio_responses(httpx_mock, transcript)

    # Mock the redacted-audio URL response
    mock_audio_file_bytes = b"pretend this is a WAV file"
    httpx_mock.add_response(
        url=REDACTED_AUDIO_URL,
        status_code=httpx.codes.OK,
        method="GET",
        content=mock_audio_file_bytes,
    )

    # Set up mocks for writing to disk
    mock_file = mocker.mock_open()
    mocker.patch("builtins.open", mock_file)

    # Download the file
    downloaded_filepath = "redacted_audio.wav"
    transcript.save_redacted_audio(downloaded_filepath)

    # Ensure correct filepath was written to
    mock_file.assert_called_once_with(downloaded_filepath, "wb")

    # Ensure correct file content was written
    write_calls = mock_file().write.call_args_list
    full_written_bytes = b"".join(call.args[0] for call in write_calls)
    assert full_written_bytes == mock_audio_file_bytes


def test_save_pii_redacted_audio_fails_if_redact_pii_not_enabled_for_transcript(
    httpx_mock: HTTPXMock,
):
    """
    Tests that an error is thrown before any requests are made if
    `redact_pii` was not enabled for the transcript and
    `save_redacted_audio` is called
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),  # standard response
        config=aai.TranscriptionConfig(),  # blank config
    )

    with pytest.raises(ValueError) as error:
        transcript.save_redacted_audio("redacted_audio.wav")

    assert "redact_pii" in str(error)

    # Ensure we never made the additional requests to get the redacted audio information
    assert len(httpx_mock.get_requests()) == 2


def test_save_pii_redacted_audio_fails_if_redact_pii_audio_not_enabled_for_transcript(
    httpx_mock: HTTPXMock,
):
    """
    Tests that an error is thrown before any requests are made if
    `redact_pii_audio` was not enabled for the transcript and
    `get_redacted_audio_url` is called
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response={
            **factories.generate_dict_factory(
                TranscriptWithPIIRedactionResponseFactory
            )(),
            "redact_pii_audio": False,
        },
        config=aai.TranscriptionConfig(
            redact_pii=True, redact_pii_policies=[aai.types.PIIRedactionPolicy.date]
        ),
    )

    with pytest.raises(ValueError) as error:
        transcript.save_redacted_audio("redacted_audio.wav")

    assert "redact_pii_audio" in str(error)

    # Ensure we never made the additional requests to get the redacted audio information
    assert len(httpx_mock.get_requests()) == 2


def test_save_pii_redacted_audio_fails_if_bad_response(httpx_mock: HTTPXMock):
    """
    Tests that `save_redacted_audio` raises a `RedactedAudioUnavailableError` if
    the request to fetch the redacted audio URL returns a `400` status code,
    indicating that the redacted audio has expired
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_audio=True,
        ),
    )

    __mock_failed_pii_audio_responses(httpx_mock, transcript)
    with pytest.raises(aai.types.RedactedAudioExpiredError):
        transcript.save_redacted_audio("redacted_audio.wav")


def test_save_pii_redacted_audio_fails_if_bad_audio_url_response(httpx_mock: HTTPXMock):
    """
    Tests that `save_redacted_audio` raises a `RedactedAudioUnavailableError` if
    the request to fetch the redacted audio **file** returns a non-200 status code
    """
    _, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            TranscriptWithPIIRedactionResponseFactory
        )(),
        config=aai.TranscriptionConfig(
            redact_pii=True,
            redact_pii_policies=[aai.types.PIIRedactionPolicy.date],
            redact_pii_audio=True,
        ),
    )

    __mock_successful_pii_audio_responses(httpx_mock, transcript)
    httpx_mock.add_response(
        url=REDACTED_AUDIO_URL,
        status_code=httpx.codes.NOT_FOUND,
        method="GET",
    )
    with pytest.raises(aai.types.RedactedAudioUnavailableError):
        transcript.save_redacted_audio("redacted_audio.wav")



import os
import sys
from importlib import reload
from unittest.mock import mock_open, patch

import httpx
import pytest
import pytest_mock
from pytest_httpx import HTTPXMock

import assemblyai as aai
from assemblyai.api import ENDPOINT_UPLOAD


class ImportFailureMocker:
    def __init__(self, module: str):
        self.module = module

    def find_spec(self, fullname, path, target=None):
        if fullname == self.module:
            raise ImportError

    def __enter__(self):
        # Remove module if already imported
        if self.module in sys.modules:
            del sys.modules[self.module]

        # Add self as first importer
        sys.meta_path.insert(0, self)
        return self

    def __exit__(self, type, value, traceback):
        # Remove self as importer
        sys.meta_path.pop(0)


def __reload_assesmblyai_module():
    reload(aai)
    aai.settings.api_key = "test"


def test_import_sdk_without_extras_installed():
    with ImportFailureMocker("pyaudio"):
        __reload_assesmblyai_module()
        # Test succeeds if no failures


def test_import_sdk_and_use_extra_functions_without_extras_installed(
    httpx_mock: HTTPXMock,
):
    with ImportFailureMocker("pyaudio"):
        __reload_assesmblyai_module()

        local_file = os.urandom(10)
        expected_upload_url = "https://example.org/audio.wav"

        # patch the reading of a local file
        with patch("builtins.open", mock_open(read_data=local_file)):
            _ = aai.extras.stream_file(filepath="audio.wav", sample_rate=44_100)

        # mock the upload endpoint
        httpx_mock.add_response(
            url=f"{aai.settings.base_url}{ENDPOINT_UPLOAD}",
            status_code=httpx.codes.OK,
            method="POST",
            json={"upload_url": expected_upload_url},
            match_content=local_file,
        )

        upload_url = aai.extras.file_from_stream(local_file)
        assert upload_url == expected_upload_url


def test_import_sdk_and_use_MicrophoneStream_without_extras_installed():
    with ImportFailureMocker("pyaudio"):
        __reload_assesmblyai_module()

        with pytest.raises(aai.extras.AssemblyAIExtrasNotInstalledError):
            aai.extras.MicrophoneStream()


def test_import_sdk_and_use_MicrophoneStream_with_extras_installed(
    mocker: pytest_mock.MockerFixture,
):
    import pyaudio

    __reload_assesmblyai_module()

    mocker.patch.object(pyaudio.PyAudio, "open", return_value=None)
    aai.extras.MicrophoneStream()

    # Test succeeds if no failures



import factory
from pytest_httpx import HTTPXMock

import tests.unit.unit_test_utils as unit_test_utils
import assemblyai as aai
from tests.unit import factories

aai.settings.api_key = "test"


class AutohighlightResultFactory(factory.Factory):
    class Meta:
        model = aai.types.AutohighlightResult

    count = factory.Faker("pyint")
    rank = factory.Faker("pyfloat")
    text = factory.Faker("sentence")
    timestamps = factory.List([factory.SubFactory(factories.TimestampFactory)])


class AutohighlightResponseFactory(factory.Factory):
    class Meta:
        model = aai.types.AutohighlightResponse

    status = aai.types.StatusResult.success
    results = factory.List([factory.SubFactory(AutohighlightResultFactory)])


class AutohighlightTranscriptResponseFactory(
    factories.TranscriptCompletedResponseFactory
):
    auto_highlights_result = factory.SubFactory(AutohighlightResponseFactory)


def test_auto_highlights_disabled_by_default(httpx_mock: HTTPXMock):
    """
    Tests that excluding `auto_highlights` from the `TranscriptionConfig` will
    result in the default behavior of it being excluded from the request body
    """
    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),
        config=aai.TranscriptionConfig(),
    )
    assert request_body.get("auto_highlights") is None
    assert transcript.auto_highlights is None


def test_auto_highlights_enabled(httpx_mock: HTTPXMock):
    """
    Tests that including `auto_highlights=True` in the `TranscriptionConfig`
    will result in `auto_highlights=True` in the request body, and that the
    response is properly parsed into a `Transcript` object
    """
    mock_response = factories.generate_dict_factory(
        AutohighlightTranscriptResponseFactory
    )()
    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=mock_response,
        config=aai.TranscriptionConfig(auto_highlights=True),
    )

    # Check that request body was properly defined
    assert request_body.get("auto_highlights") is True

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None
    assert transcript.auto_highlights is not None
    assert (
        transcript.auto_highlights.status
        == mock_response["auto_highlights_result"]["status"]
    )

    assert transcript.auto_highlights.results is not None
    assert len(transcript.auto_highlights.results) > 0
    assert len(transcript.auto_highlights.results) == len(
        mock_response["auto_highlights_result"]["results"]
    )

    for response_result, transcript_result in zip(
        mock_response["auto_highlights_result"]["results"],
        transcript.auto_highlights.results,
    ):
        assert transcript_result.count == response_result["count"]
        assert transcript_result.rank == response_result["rank"]
        assert transcript_result.text == response_result["text"]

        for response_timestamp, transcript_timestamp in zip(
            response_result["timestamps"], transcript_result.timestamps
        ):
            assert transcript_timestamp.start == response_timestamp["start"]
            assert transcript_timestamp.end == response_timestamp["end"]



import json
from typing import Any, Dict, Tuple

import httpx
from pytest_httpx import HTTPXMock

import assemblyai as aai
from assemblyai.api import ENDPOINT_TRANSCRIPT
from tests.unit import factories


def submit_mock_transcription_request(
    httpx_mock: HTTPXMock,
    mock_response: Dict[str, Any],
    config: aai.TranscriptionConfig,
) -> Tuple[Dict[str, Any], aai.transcriber.Transcript]:
    """
    Helper function to abstract calling transcriber with given parameters,
    and perform some common assertions.

    Args:
        httpx_mock: HTTPXMock instance to use for mocking requests
        mock_response: Dict to use as mock response from API
        config: The `TranscriptionConfig` to use for transcription

    Returns:
        A tuple containing the JSON body of the initial submission request,
        and the `Transcript` object parsed from the mock response
    """

    mock_transcript_id = mock_response.get("id", "mock_id")

    # Mock initial submission response (transcript is processing)
    mock_processing_response = factories.generate_dict_factory(
        factories.TranscriptProcessingResponseFactory
    )()

    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_TRANSCRIPT}",
        status_code=httpx.codes.OK,
        method="POST",
        json={
            **mock_processing_response,
            "id": mock_transcript_id,  # inject ID from main mock response
        },
    )

    # Mock polling-for-completeness response, with completed transcript
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_TRANSCRIPT}/{mock_transcript_id}",
        status_code=httpx.codes.OK,
        method="GET",
        json=mock_response,
    )

    # == Make API request via SDK ==
    transcript = aai.Transcriber().transcribe(
        data="https://example.org/audio.wav",
        config=config,
    )

    # Check that submission and polling requests were made
    assert len(httpx_mock.get_requests()) == 2

    # Extract body of initial submission request
    request = httpx_mock.get_requests()[0]
    request_body = json.loads(request.content.decode())

    return request_body, transcript



"""
Contains factories that are used for mocking certain requests/responses
from AssemblyAI's API.
"""

from enum import Enum
from functools import partial
from typing import Any, Callable, Dict

import factory
import factory.base

import assemblyai as aai
from assemblyai import types


class TimestampFactory(factory.Factory):
    class Meta:
        model = aai.Timestamp

    start = factory.Faker("pyint")
    end = factory.Faker("pyint")


class WordFactory(factory.Factory):
    class Meta:
        model = aai.Word

    text = factory.Faker("word")
    start = factory.Faker("pyint")
    end = factory.Faker("pyint")
    confidence = factory.Faker("pyfloat", min_value=0.0, max_value=1.0)


class UtteranceWordFactory(WordFactory):
    class Meta:
        model = aai.UtteranceWord

    speaker = factory.Faker("name")


class UtteranceFactory(UtteranceWordFactory):
    class Meta:
        model = aai.Utterance

    words = factory.List([factory.SubFactory(UtteranceWordFactory)])


class ChapterFactory(factory.Factory):
    class Meta:
        model = types.Chapter

    summary = factory.Faker("sentence")
    headline = factory.Faker("sentence")
    gist = factory.Faker("sentence")
    start = factory.Faker("pyint")
    end = factory.Faker("pyint")


class BaseTranscriptFactory(factory.Factory):
    class Meta:
        model = types.BaseTranscript

    language_code = "en"
    audio_url = factory.Faker("url")
    punctuate = True
    format_text = True
    dual_channel = True
    webhook_url = None
    webhook_auth_header_name = None
    audio_start_from = None
    audio_end_at = None
    word_boost = None
    boost_param = None
    filter_profanity = False
    redact_pii = False
    redact_pii_audio = False
    redact_pii_policies = None
    redact_pii_sub = None
    speaker_labels = False
    content_safety = False
    iab_categories = False
    custom_spelling = None
    disfluencies = False
    sentiment_analysis = False
    auto_chapters = False
    entity_detection = False
    summarization = False
    summary_model = None
    summary_type = None
    auto_highlights = False
    language_detection = False
    speech_threshold = None


class BaseTranscriptResponseFactory(BaseTranscriptFactory):
    class Meta:
        model = types.TranscriptResponse

    id = factory.Faker("uuid4")
    status = aai.TranscriptStatus.completed
    error = None
    text = factory.Faker("text")
    words = factory.List([factory.SubFactory(WordFactory)])
    utterances = factory.List([factory.SubFactory(UtteranceFactory)])
    confidence = factory.Faker("pyfloat", min_value=0.0, max_value=1.0)
    audio_duration = factory.Faker("pyint")
    webhook_auth = False
    webhook_status_code = None


class TranscriptDeletedResponseFactory(BaseTranscriptResponseFactory):
    language_code = None
    audio_url = "http://deleted_by_user"
    text = "Deleted by user."
    words = None
    utterances = None
    confidence = None
    punctuate = None
    format_text = None
    dual_channel = None
    webhook_url = "http://deleted_by_user"
    webhook_status_code = None
    webhook_auth = False
    # webhook_auth_header_name = None  # not yet supported in SDK
    speed_boost = None
    auto_highlights = None
    audio_start_from = None
    audio_end_at = None
    word_boost = None
    boost_param = None
    filter_profanity = None
    redact_pii_audio = None
    # redact_pii_quality = None # not yet supported in SDK
    redact_pii_policies = None
    redact_pii_sub = None
    speaker_labels = None
    error = None
    content_safety = None
    iab_categories = None
    content_safety_labels = None
    iab_categories = None
    language_detection = None
    custom_spelling = None
    # cluster_id = None  # not yet supported in SDK
    # custom_topics = None  # not yet supported in SDK
    # topics = None  # not yet supported in SDK
    speech_threshold = None
    chapters = None
    entities = None
    speakers_expected = None
    summary = None
    sentiment_analysis = None


class TranscriptCompletedResponseFactory(BaseTranscriptResponseFactory):
    pass


class TranscriptCompletedResponseFactoryBest(BaseTranscriptResponseFactory):
    speech_model = "best"


class TranscriptCompletedResponseFactoryNano(BaseTranscriptResponseFactory):
    speech_model = "nano"


class TranscriptQueuedResponseFactory(BaseTranscriptFactory):
    class Meta:
        model = types.TranscriptResponse

    id = factory.Faker("uuid4")
    status = aai.TranscriptStatus.queued
    text = None
    words = None
    utterances = None
    confidence = None
    audio_duration = None


class TranscriptProcessingResponseFactory(BaseTranscriptFactory):
    class Meta:
        model = types.TranscriptResponse

    id = factory.Faker("uuid4")
    status = aai.TranscriptStatus.processing
    text = None
    words = None
    utterances = None
    confidence = None
    audio_duration = None


class TranscriptErrorResponseFactory(BaseTranscriptFactory):
    class Meta:
        model = types.TranscriptResponse

    status = aai.TranscriptStatus.error
    error = "Aw, snap!"


class TranscriptRequestFactory(BaseTranscriptFactory):
    class Meta:
        model = types.TranscriptRequest


class PageDetails(factory.Factory):
    class Meta:
        model = types.PageDetails

    current_url = factory.Faker("url")
    limit = 10
    next_url = None
    prev_url = None
    result_count = 2


class TranscriptItem(factory.Factory):
    class Meta:
        model = types.TranscriptItem

    audio_url = factory.Faker("url")
    created = factory.Faker("iso8601")
    id = factory.Faker("uuid4")
    resource_url = factory.Faker("url")
    status = aai.TranscriptStatus.completed
    completed = None
    error = None


class ListTranscriptResponse(factory.Factory):
    class Meta:
        model = types.ListTranscriptResponse

    page_details = factory.SubFactory(PageDetails)
    transcripts = factory.List(
        [
            factory.SubFactory(TranscriptItem),
            factory.SubFactory(TranscriptItem),
        ]
    )


class LemurUsage(factory.Factory):
    class Meta:
        model = types.LemurUsage

    input_tokens = factory.Faker("pyint")
    output_tokens = factory.Faker("pyint")


class LemurQuestionAnswer(factory.Factory):
    class Meta:
        model = types.LemurQuestionAnswer

    question = factory.Faker("text")
    answer = factory.Faker("text")


class LemurQuestionResponse(factory.Factory):
    class Meta:
        model = types.LemurQuestionResponse

    request_id = factory.Faker("uuid4")
    usage = factory.SubFactory(LemurUsage)
    response = factory.List(
        [
            factory.SubFactory(LemurQuestionAnswer),
            factory.SubFactory(LemurQuestionAnswer),
        ]
    )


class LemurSummaryResponse(factory.Factory):
    class Meta:
        model = types.LemurSummaryResponse

    request_id = factory.Faker("uuid4")
    usage = factory.SubFactory(LemurUsage)
    response = factory.Faker("text")


class LemurActionItemsResponse(factory.Factory):
    class Meta:
        model = types.LemurActionItemsResponse

    request_id = factory.Faker("uuid4")
    usage = factory.SubFactory(LemurUsage)
    response = factory.Faker("text")


class LemurTaskResponse(factory.Factory):
    class Meta:
        model = types.LemurTaskResponse

    request_id = factory.Faker("uuid4")
    usage = factory.SubFactory(LemurUsage)
    response = factory.Faker("text")


class LemurStringResponse(factory.Factory):
    class Meta:
        model = types.LemurStringResponse

    request_id = factory.Faker("uuid4")
    usage = factory.SubFactory(LemurUsage)
    response = factory.Faker("text")


class LemurPurgeResponse(factory.Factory):
    class Meta:
        model = types.LemurPurgeResponse

    request_id = factory.Faker("uuid4")
    request_id_to_purge = factory.Faker("uuid4")
    deleted = True


class WordSearchMatchFactory(factory.Factory):
    class Meta:
        model = types.WordSearchMatch

    text = factory.Faker("text")
    count = factory.Faker("pyint")
    timestamps = [(123, 456)]
    indexes = [123, 456]


class WordSearchMatchResponseFactory(factory.Factory):
    class Meta:
        model = types.WordSearchMatchResponse

    total_count = factory.Faker("pyint")

    matches = factory.List([factory.SubFactory(WordSearchMatchFactory)])


class SentenceFactory(WordFactory):
    class Meta:
        model = types.Sentence

    words = factory.List([factory.SubFactory(WordFactory)])


class ParagraphFactory(SentenceFactory):
    class Meta:
        model = types.Paragraph


class SentencesResponseFactory(factory.Factory):
    class Meta:
        model = types.SentencesResponse

    sentences = factory.List([factory.SubFactory(SentenceFactory)])
    confidence = factory.Faker("pyfloat", min_value=0.0, max_value=1.0)
    audio_duration = factory.Faker("pyint")


class ParagraphsResponseFactory(factory.Factory):
    class Meta:
        model = types.ParagraphsResponse

    paragraphs = factory.List([factory.SubFactory(ParagraphFactory)])
    confidence = factory.Faker("pyfloat", min_value=0.0, max_value=1.0)
    audio_duration = factory.Faker("pyint")


def generate_dict_factory(f: factory.Factory) -> Callable[[], Dict[str, Any]]:
    """
    Creates a dict factory from the given *Factory class.

    Args:
        f: The factory to create a dict factory from.
    """

    def stub_is_list(stub: factory.base.StubObject) -> bool:
        try:
            return all(k.isdigit() for k in stub.__dict__.keys())
        except AttributeError:
            return False

    def convert_dict_from_stub(stub: factory.base.StubObject) -> Dict[str, Any]:
        stub_dict = stub.__dict__
        for key, value in stub_dict.items():
            if isinstance(value, factory.base.StubObject):
                stub_dict[key] = (
                    [convert_dict_from_stub(v) for v in value.__dict__.values()]
                    if stub_is_list(value)
                    else convert_dict_from_stub(value)
                )
            elif isinstance(value, Enum):
                stub_dict[key] = value.value
        return stub_dict

    def dict_factory(f, **kwargs):
        stub = f.stub(**kwargs)
        stub_dict = convert_dict_from_stub(stub)
        return stub_dict

    return partial(dict_factory, f)



import assemblyai as aai


def test_reset_client_on_settings_change():
    """
    Test that the settings are reset when the global settings have changed.
    """
    aai.settings.api_key = "before"
    transcriber = aai.Transcriber()

    assert transcriber._client.settings.api_key == "before"

    # Reset it to "test" again. All other tests are also working with this value
    aai.settings.api_key = "test"
    transcriber = aai.Transcriber()

    assert transcriber._client.settings.api_key == "test"



import factory
from pytest_httpx import HTTPXMock

import tests.unit.unit_test_utils as unit_test_utils
import assemblyai as aai
from tests.unit import factories

aai.settings.api_key = "test"


class IABLabelResultFactory(factory.Factory):
    class Meta:
        model = aai.types.IABLabelResult

    relevance = factory.Faker("pyfloat", min_value=0, max_value=1)
    label = factory.Faker("word")


class IABResultFactory(factory.Factory):
    class Meta:
        model = aai.types.IABResult

    text = factory.Faker("sentence")
    labels = factory.List([factory.SubFactory(IABLabelResultFactory)])
    timestamp = factory.SubFactory(factories.TimestampFactory)


class IABResponseFactory(factory.Factory):
    class Meta:
        model = aai.types.IABResponse

    status = aai.types.StatusResult.success.value
    results = factory.List([factory.SubFactory(IABResultFactory)])
    summary = factory.Dict(
        {
            "Automotive>AutoType>ConceptCars": factory.Faker(
                "pyfloat", min_value=0, max_value=1
            )
        }
    )


class IABCategoriesResponseFactory(factories.TranscriptCompletedResponseFactory):
    iab_categories_result = factory.SubFactory(IABResponseFactory)


def test_iab_categories_disabled_by_default(httpx_mock: HTTPXMock):
    """
    Tests that excluding `iab_categories` from the `TranscriptionConfig` will
    result in the default behavior of it being excluded from the request body
    """

    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),
        config=aai.TranscriptionConfig(),
    )
    assert request_body.get("iab_categories") is None
    assert transcript.iab_categories is None


def test_iab_categories_enabled(httpx_mock: HTTPXMock):
    """
    Tests that including `iab_categories=True` in the `TranscriptionConfig` will
    result in `iab_categories` being included in the request body, and that
    the response will be properly parsed into the `Transcript` object
    """

    mock_response = factories.generate_dict_factory(IABCategoriesResponseFactory)()

    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=mock_response,
        config=aai.TranscriptionConfig(iab_categories=True),
    )

    assert request_body.get("iab_categories") is True

    assert transcript.error is None

    assert transcript.iab_categories is not None
    assert transcript.iab_categories.status == mock_response.get(
        "iab_categories_result", {}
    ).get("status")

    # Validate results
    response_results = mock_response.get("iab_categories_result", {}).get("results", [])
    transcript_results = transcript.iab_categories.results

    assert transcript_results is not None
    assert len(transcript_results) == len(response_results)
    assert len(transcript_results) > 0

    for response_result, transcript_result in zip(response_results, transcript_results):
        assert transcript_result.text == response_result.get("text")
        assert len(transcript_result.text) > 0

        assert len(transcript_result.labels) > 0
        assert len(transcript_result.labels) == len(response_result.get("labels", []))
        for response_label, transcript_label in zip(
            response_result.get("labels", []), transcript_result.labels
        ):
            assert transcript_label.relevance == response_label.get("relevance")
            assert transcript_label.label == response_label.get("label")

    # Validate summary
    response_summary = mock_response.get("iab_categories_result", {}).get("summary", {})
    transcript_summary = transcript.iab_categories.summary

    assert transcript_summary is not None
    assert len(transcript_summary) > 0
    assert transcript_summary == response_summary



from . import extras
from .__version__ import __version__
from .client import Client
from .lemur import Lemur
from .transcriber import RealtimeTranscriber, Transcriber, Transcript, TranscriptGroup
from .types import (
    AssemblyAIError,
    AudioEncoding,
    AutohighlightResponse,
    AutohighlightResult,
    Chapter,
    ContentSafetyLabel,
    ContentSafetyLabelResult,
    ContentSafetyResponse,
    ContentSafetySeverityScore,
    Entity,
    EntityType,
    IABLabelResult,
    IABResponse,
    IABResult,
    LanguageCode,
    LemurActionItemsResponse,
    LemurError,
    LemurModel,
    LemurPurgeRequest,
    LemurPurgeResponse,
    LemurQuestion,
    LemurQuestionAnswer,
    LemurQuestionResponse,
    LemurSource,
    LemurSourceType,
    LemurStringResponse,
    LemurSummaryResponse,
    LemurTaskResponse,
    LemurTranscriptSource,
    LemurUsage,
    ListTranscriptParameters,
    ListTranscriptResponse,
    PageDetails,
    Paragraph,
    PIIRedactedAudioQuality,
    PIIRedactionPolicy,
    PIISubstitutionPolicy,
    RawTranscriptionConfig,
    RealtimeError,
    RealtimeFinalTranscript,
    RealtimePartialTranscript,
    RealtimeSessionInformation,
    RealtimeSessionOpened,
    RealtimeTranscript,
    RealtimeWord,
    Sentence,
    Sentiment,
    SentimentType,
    Settings,
    SpeechModel,
    StatusResult,
    SummarizationModel,
    SummarizationType,
    Timestamp,
    TranscriptError,
    TranscriptionConfig,
    TranscriptItem,
    TranscriptStatus,
    Utterance,
    UtteranceWord,
    Word,
    WordBoost,
    WordSearchMatch,
)

settings = Settings()
"""Global settings object that applies to all classes that use the `Client` class."""


__all__ = [
    # types
    "AssemblyAIError",
    "AudioEncoding",
    "AutohighlightResponse",
    "AutohighlightResult",
    "Chapter",
    "Client",
    "ContentSafetyLabel",
    "ContentSafetyLabelResult",
    "ContentSafetyResponse",
    "ContentSafetySeverityScore",
    "Entity",
    "EntityType",
    "IABLabelResult",
    "IABResponse",
    "IABResult",
    "LanguageCode",
    "Lemur",
    "LemurActionItemsResponse",
    "LemurError",
    "LemurModel",
    "LemurPurgeRequest",
    "LemurPurgeResponse",
    "LemurSource",
    "LemurSourceType",
    "LemurTranscriptSource",
    "LemurQuestion",
    "LemurQuestionAnswer",
    "LemurQuestionResponse",
    "LemurStringResponse",
    "LemurSummaryResponse",
    "LemurTaskResponse",
    "LemurUsage",
    "ListTranscriptParameters",
    "ListTranscriptResponse",
    "PageDetails",
    "Sentence",
    "Sentiment",
    "SentimentType",
    "Settings",
    "SpeechModel",
    "StatusResult",
    "SummarizationModel",
    "SummarizationType",
    "Timestamp",
    "Transcriber",
    "TranscriptionConfig",
    "Transcript",
    "TranscriptError",
    "TranscriptGroup",
    "TranscriptItem",
    "TranscriptStatus",
    "Utterance",
    "UtteranceWord",
    "Paragraph",
    "PIIRedactedAudioQuality",
    "PIISubstitutionPolicy",
    "PIIRedactionPolicy",
    "RawTranscriptionConfig",
    "Word",
    "WordBoost",
    "WordSearchMatch",
    "RealtimeTranscriber",
    "RealtimeError",
    "RealtimeFinalTranscript",
    "RealtimePartialTranscript",
    "RealtimeSessionInformation",
    "RealtimeSessionOpened",
    "RealtimeTranscript",
    "RealtimeWord",
    # package globals
    "settings",
    # packages
    "extras",
    # version
    "__version__",
]



import uuid

import httpx
from pytest_httpx import HTTPXMock

import assemblyai as aai
from assemblyai.api import ENDPOINT_TRANSCRIPT
from tests.unit import factories


def test_transcript_group_accepts_transcript_ids():
    """
    Tests whether a TranscriptGroup accepts transcript IDs.
    """
    transcript_ids = [str(uuid.uuid4()), str(uuid.uuid4())]

    transcript_group = aai.TranscriptGroup(transcript_ids=transcript_ids)

    assert [transcript.id for transcript in transcript_group] == transcript_ids


def test_transcript_group_check_status():
    """
    Tests the TranscriptGroup's status
    """

    # create a mock response of a completed transcript
    mock_completed_transcript = factories.generate_dict_factory(
        factories.TranscriptCompletedResponseFactory
    )()

    mock_queued_transcript = factories.generate_dict_factory(
        factories.TranscriptQueuedResponseFactory
    )()

    mock_processing_transcript = factories.generate_dict_factory(
        factories.TranscriptProcessingResponseFactory
    )()

    mock_error_transcript = factories.generate_dict_factory(
        factories.TranscriptErrorResponseFactory
    )()

    transcript_completed = aai.Transcript.from_response(
        client=aai.Client.get_default(),
        response=aai.types.TranscriptResponse(**mock_completed_transcript),
    )

    transcript_queued = aai.Transcript.from_response(
        client=aai.Client.get_default(),
        response=aai.types.TranscriptResponse(**mock_queued_transcript),
    )

    transcript_processing = aai.Transcript.from_response(
        client=aai.Client.get_default(),
        response=aai.types.TranscriptResponse(**mock_processing_transcript),
    )

    transcript_error = aai.Transcript.from_response(
        client=aai.Client.get_default(),
        response=aai.types.TranscriptResponse(**mock_error_transcript),
    )

    transcript_group = aai.TranscriptGroup()

    transcript_group.add_transcript(transcript_completed)
    assert transcript_group.status == aai.TranscriptStatus.completed

    transcript_group.add_transcript(transcript_queued)
    assert transcript_group.status == aai.TranscriptStatus.queued

    transcript_group.add_transcript(transcript_processing)
    assert transcript_group.status == aai.TranscriptStatus.queued

    transcript_group.add_transcript(transcript_error)
    assert transcript_group.status == aai.TranscriptStatus.error


def test_get_by_ids(httpx_mock: HTTPXMock):
    transcript_ids = ["123", "456"]
    mock_transcript_response = factories.generate_dict_factory(
        factories.TranscriptCompletedResponseFactory
    )()
    for transcript_id in transcript_ids:
        httpx_mock.add_response(
            url=f"{aai.settings.base_url}{ENDPOINT_TRANSCRIPT}/{transcript_id}",
            status_code=httpx.codes.OK,
            method="GET",
            json=mock_transcript_response,
        )

    transcript_group = aai.TranscriptGroup.get_by_ids(transcript_ids)

    assert isinstance(transcript_group, aai.TranscriptGroup)
    assert transcript_group.status == aai.TranscriptStatus.completed
    for transcript in transcript_group:
        assert transcript.id in transcript_ids
        transcript_ids.remove(transcript.id)

        assert transcript.error is None
    assert len(transcript_ids) == 0


def test_get_by_id_async(httpx_mock: HTTPXMock):
    transcript_ids = ["123", "456"]
    mock_transcript_response = factories.generate_dict_factory(
        factories.TranscriptCompletedResponseFactory
    )()
    for transcript_id in transcript_ids:
        httpx_mock.add_response(
            url=f"{aai.settings.base_url}{ENDPOINT_TRANSCRIPT}/{transcript_id}",
            status_code=httpx.codes.OK,
            method="GET",
            json=mock_transcript_response,
        )

    transcript_group_future = aai.TranscriptGroup.get_by_ids_async(transcript_ids)
    transcript_group = transcript_group_future.result()

    assert isinstance(transcript_group, aai.TranscriptGroup)
    assert transcript_group.status == aai.TranscriptStatus.completed
    for transcript in transcript_group:
        assert transcript.id in transcript_ids
        transcript_ids.remove(transcript.id)

        assert transcript.error is None
    assert len(transcript_ids) == 0



import factory
from pytest_httpx import HTTPXMock

import tests.unit.unit_test_utils as unit_test_utils
import assemblyai as aai
from tests.unit import factories

aai.settings.api_key = "test"


class SentimentFactory(factories.WordFactory):
    sentiment = factory.Faker("enum", enum_cls=aai.types.SentimentType)
    speaker = factory.Faker("name")


class SentimentAnalysisResponseFactory(factories.TranscriptCompletedResponseFactory):
    sentiment_analysis_results = factory.List([factory.SubFactory(SentimentFactory)])


def test_sentiment_analysis_disabled_by_default(httpx_mock: HTTPXMock):
    """
    Tests that excluding `sentiment_analysis` from the `TranscriptionConfig` will
    result in the default behavior of it being excluded from the request body
    """
    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=factories.generate_dict_factory(
            factories.TranscriptCompletedResponseFactory
        )(),
        config=aai.TranscriptionConfig(),
    )
    assert request_body.get("sentiment_analysis") is None
    assert transcript.sentiment_analysis is None


def test_sentiment_analysis_enabled(httpx_mock: HTTPXMock):
    """
    Tests that including `sentiment_analysis=True` in the `TranscriptionConfig`
    will result in `sentiment_analysis=True` in the request body, and that the
    response is properly parsed into a `Transcript` object
    """
    mock_response = factories.generate_dict_factory(SentimentAnalysisResponseFactory)()
    request_body, transcript = unit_test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response=mock_response,
        config=aai.TranscriptionConfig(sentiment_analysis=True),
    )

    # Check that request body was properly defined
    assert request_body.get("sentiment_analysis") is True

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None

    assert transcript.sentiment_analysis is not None
    assert len(transcript.sentiment_analysis) > 0
    assert len(transcript.sentiment_analysis) == len(
        mock_response["sentiment_analysis_results"]
    )

    for response_sentiment_result, transcript_sentiment_result in zip(
        mock_response["sentiment_analysis_results"],
        transcript.sentiment_analysis,
    ):
        assert transcript_sentiment_result.text == response_sentiment_result["text"]
        assert transcript_sentiment_result.start == response_sentiment_result["start"]
        assert transcript_sentiment_result.end == response_sentiment_result["end"]
        assert (
            transcript_sentiment_result.confidence
            == response_sentiment_result["confidence"]
        )
        assert (
            transcript_sentiment_result.sentiment.value
            == response_sentiment_result["sentiment"]
        )
        assert (
            transcript_sentiment_result.speaker == response_sentiment_result["speaker"]
        )



import datetime
import json
import uuid
from unittest.mock import MagicMock
from urllib.parse import urlencode

import httpx
import pytest
import websockets.exceptions
from faker import Faker
from pytest_httpx import HTTPXMock
from pytest_mock import MockFixture

import assemblyai as aai
from assemblyai.api import ENDPOINT_REALTIME_TOKEN

aai.settings.api_key = "test"


def _disable_rw_threads(mocker: MockFixture):
    """
    Disable the read/write threads for the websocket
    """

    mocker.patch("threading.Thread.start", return_value=None)


@pytest.mark.parametrize(
    "encoding,token,expected_header",
    [
        (None, None, {"Authorization": "test"}),
        (aai.AudioEncoding.pcm_s16le, None, {"Authorization": "test"}),
        (aai.AudioEncoding.pcm_mulaw, None, {"Authorization": "test"}),
        (None, "12345678", None),
        (aai.AudioEncoding.pcm_s16le, "12345678", None),
    ],
)
def test_realtime_connect_has_parameters(
    encoding, token, expected_header, mocker: MockFixture
):
    """
    Test that the connect method has the correct parameters set
    """
    aai.settings.base_url = "https://api.assemblyai.com"

    actual_url = None
    actual_additional_headers = None
    actual_open_timeout = None

    def mocked_websocket_connect(
        url: str, additional_headers: dict, open_timeout: float
    ):
        nonlocal actual_url, actual_additional_headers, actual_open_timeout
        actual_url = url
        actual_additional_headers = additional_headers
        actual_open_timeout = open_timeout

    mocker.patch(
        "assemblyai.transcriber.websocket_connect",
        new=mocked_websocket_connect,
    )
    _disable_rw_threads(mocker)

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda: None,
        on_error=lambda error: print(error),
        sample_rate=44_100,
        word_boost=["AssemblyAI"],
        encoding=encoding,
        token=token,
    )

    transcriber.connect(timeout=15.0)

    params = dict(sample_rate=44100, word_boost=json.dumps(["AssemblyAI"]))
    if encoding:
        params["encoding"] = encoding.value
    if token:
        params["token"] = token

    assert actual_url == f"wss://api.assemblyai.com/v2/realtime/ws?{urlencode(params)}"
    assert actual_additional_headers == expected_header
    assert actual_open_timeout == 15.0


def test_realtime_connect_succeeds(mocker: MockFixture):
    """
    Tests that the `RealtimeTranscriber` successfully connects to the `real-time` service.
    """
    on_error_called = False

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=on_error,
        sample_rate=44_100,
    )

    mocker.patch(
        "assemblyai.transcriber.websocket_connect",
        return_value=MagicMock(),
    )

    # mock the read/write threads
    _disable_rw_threads(mocker)

    # should pass
    transcriber.connect()

    # no errors should be called
    assert not on_error_called


def test_realtime_token_connect_succeeds(mocker: MockFixture):
    """
    Tests that the `RealtimeTranscriber` successfully connects
    to the `real-time` service when a token is used.
    """
    on_error_called = False

    # reset the API key
    mocker.patch("assemblyai.settings.api_key", new=None)

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None, on_error=on_error, sample_rate=44_100, token="12345"
    )

    mocker.patch(
        "assemblyai.transcriber.websocket_connect",
        return_value=MagicMock(),
    )

    # mock the read/write threads
    _disable_rw_threads(mocker)

    # should pass
    transcriber.connect()

    # no errors should be called
    assert not on_error_called


def test_realtime_connect_fails(mocker: MockFixture):
    """
    Tests that the `RealtimeTranscriber` fails to connect to the `real-time` service.
    """

    on_error_called = False

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True

        assert isinstance(error, aai.RealtimeError)
        assert "connection failed" in str(error)

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=on_error,
        sample_rate=44_100,
    )
    mocker.patch(
        "assemblyai.transcriber.websocket_connect",
        side_effect=Exception("connection failed"),
    )

    transcriber.connect()

    assert on_error_called


def test_realtime__read_succeeds(mocker: MockFixture, faker: Faker):
    """
    Tests the `_read` method of the `_RealtimeTranscriberImpl` class.
    """

    expected_transcripts = [
        aai.RealtimeFinalTranscript(
            created=faker.date_time(),
            text=faker.sentence(),
            audio_start=0,
            audio_end=1,
            confidence=1.0,
            words=[],
            punctuated=True,
            text_formatted=True,
        )
    ]

    received_transcripts = []

    def on_data(data: aai.RealtimeTranscript):
        nonlocal received_transcripts
        received_transcripts.append(data)

    transcriber = aai.RealtimeTranscriber(
        on_data=on_data,
        on_error=lambda _: None,
        sample_rate=44_100,
    )

    transcriber._impl._websocket = MagicMock()
    websocket_recv = [
        json.dumps(msg.dict(), default=str) for msg in expected_transcripts
    ]
    transcriber._impl._websocket.recv.side_effect = websocket_recv

    with pytest.raises(StopIteration):
        transcriber._impl._read()

    assert received_transcripts == expected_transcripts


def test_realtime__read_fails(mocker: MockFixture):
    """
    Tests the `_read` method of the `_RealtimeTranscriberImpl` class.
    """

    on_error_called = False

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=on_error,
        sample_rate=44_100,
    )

    transcriber._impl._websocket = MagicMock()
    error = websockets.exceptions.ConnectionClosedOK(rcvd=None, sent=None)
    transcriber._impl._websocket.recv.side_effect = error

    transcriber._impl._read()

    assert on_error_called


def test_realtime__write_succeeds(mocker: MockFixture):
    """
    Tests the `_write` method of the `_RealtimeTranscriberImpl` class.
    """
    audio_chunks = [
        bytes([1, 2, 3, 4, 5]),
        bytes([6, 7, 8, 9, 10]),
    ]

    actual_sent = []

    def mocked_send(data: str):
        nonlocal actual_sent
        actual_sent.append(data)

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=lambda _: None,
        sample_rate=44_100,
    )

    transcriber._impl._websocket = MagicMock()
    transcriber._impl._websocket.send = mocked_send
    transcriber._impl._stop_event.is_set = MagicMock(side_effect=[False, False, True])

    transcriber.stream(audio_chunks[0])
    transcriber.stream(audio_chunks[1])

    transcriber._impl._write()

    # assert that the correct data was sent (= the exact input bytes)
    assert len(actual_sent) == 2
    assert actual_sent[0] == audio_chunks[0]
    assert actual_sent[1] == audio_chunks[1]


def test_realtime__handle_message_session_begins(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with the `SessionBegins` message.
    """

    test_message = {
        "message_type": "SessionBegins",
        "session_id": str(uuid.uuid4()),
        "expires_at": datetime.datetime.now().isoformat(),
    }

    on_open_called = False

    def on_open(session_opened: aai.RealtimeSessionOpened):
        nonlocal on_open_called
        on_open_called = True
        assert isinstance(session_opened, aai.RealtimeSessionOpened)
        assert session_opened.session_id == uuid.UUID(test_message["session_id"])
        assert session_opened.expires_at.isoformat() == test_message["expires_at"]

    transcriber = aai.RealtimeTranscriber(
        on_open=on_open,
        on_data=lambda _: None,
        on_error=lambda _: None,
        sample_rate=44_100,
    )

    transcriber._impl._handle_message(test_message)

    assert on_open_called


def test_realtime__handle_message_partial_transcript(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with the `PartialTranscript` message.
    """

    test_message = {
        "message_type": "PartialTranscript",
        "text": "hello world",
        "audio_start": 0,
        "audio_end": 1500,
        "confidence": 0.99,
        "created": datetime.datetime.now().isoformat(),
        "words": [
            {
                "text": "hello",
                "start": 0,
                "end": 500,
                "confidence": 0.99,
            },
            {
                "text": "world",
                "start": 500,
                "end": 1500,
                "confidence": 0.99,
            },
        ],
    }

    on_data_called = False

    def on_data(data: aai.RealtimePartialTranscript):
        nonlocal on_data_called
        on_data_called = True
        assert isinstance(data, aai.RealtimePartialTranscript)
        assert data.text == test_message["text"]
        assert data.audio_start == test_message["audio_start"]
        assert data.audio_end == test_message["audio_end"]
        assert data.confidence == test_message["confidence"]
        assert data.created.isoformat() == test_message["created"]
        assert data.words == [
            aai.RealtimeWord(
                text=test_message["words"][0]["text"],
                start=test_message["words"][0]["start"],
                end=test_message["words"][0]["end"],
                confidence=test_message["words"][0]["confidence"],
            ),
            aai.RealtimeWord(
                text=test_message["words"][1]["text"],
                start=test_message["words"][1]["start"],
                end=test_message["words"][1]["end"],
                confidence=test_message["words"][1]["confidence"],
            ),
        ]

    transcriber = aai.RealtimeTranscriber(
        on_data=on_data,
        on_error=lambda _: None,
        sample_rate=44_100,
    )

    transcriber._impl._handle_message(test_message)

    assert on_data_called


def test_realtime__handle_message_final_transcript(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with the `FinalTranscript` message.
    """

    test_message = {
        "message_type": "FinalTranscript",
        "text": "Hello, world!",
        "audio_start": 0,
        "audio_end": 1500,
        "confidence": 0.99,
        "created": datetime.datetime.now().isoformat(),
        "punctuated": True,
        "text_formatted": True,
        "words": [
            {
                "text": "Hello,",
                "start": 0,
                "end": 500,
                "confidence": 0.99,
            },
            {
                "text": "world!",
                "start": 500,
                "end": 1500,
                "confidence": 0.99,
            },
        ],
    }

    on_data_called = False

    def on_data(data: aai.RealtimeFinalTranscript):
        nonlocal on_data_called
        on_data_called = True
        assert isinstance(data, aai.RealtimeFinalTranscript)
        assert data.text == test_message["text"]
        assert data.audio_start == test_message["audio_start"]
        assert data.audio_end == test_message["audio_end"]
        assert data.confidence == test_message["confidence"]
        assert data.created.isoformat() == test_message["created"]
        assert data.punctuated == test_message["punctuated"]
        assert data.text_formatted == test_message["text_formatted"]
        assert data.words == [
            aai.RealtimeWord(
                text=test_message["words"][0]["text"],
                start=test_message["words"][0]["start"],
                end=test_message["words"][0]["end"],
                confidence=test_message["words"][0]["confidence"],
            ),
            aai.RealtimeWord(
                text=test_message["words"][1]["text"],
                start=test_message["words"][1]["start"],
                end=test_message["words"][1]["end"],
                confidence=test_message["words"][1]["confidence"],
            ),
        ]

    transcriber = aai.RealtimeTranscriber(
        on_data=on_data,
        on_error=lambda _: None,
        sample_rate=44_100,
    )

    transcriber._impl._handle_message(test_message)

    assert on_data_called


def test_realtime__handle_message_error_message(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with the error message.
    """

    test_message = {
        "error": "test error",
    }

    on_error_called = False

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True
        assert isinstance(error, aai.RealtimeError)
        assert str(error) == test_message["error"]

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=on_error,
        sample_rate=44_100,
    )

    transcriber._impl._handle_message(test_message)

    assert on_error_called


def test_realtime__handle_message_session_information_message(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with the session information message.
    """

    test_message = {
        "message_type": "SessionInformation",
        "audio_duration_seconds": 3000.0,
    }

    on_extra_session_information_called = False

    def on_extra_session_information(data: aai.RealtimeSessionInformation):
        nonlocal on_extra_session_information_called
        on_extra_session_information_called = True
        assert isinstance(data, aai.RealtimeSessionInformation)
        assert data.audio_duration_seconds == test_message["audio_duration_seconds"]

    transcriber = aai.RealtimeTranscriber(
        on_data=lambda _: None,
        on_error=lambda _: None,
        sample_rate=44_100,
        on_extra_session_information=on_extra_session_information,
    )

    transcriber._impl._handle_message(test_message)

    assert on_extra_session_information_called


def test_realtime__handle_message_unknown_message(mocker: MockFixture):
    """
    Tests the `_handle_message` method of the `_RealtimeTranscriberImpl` class
    with an unknown message.
    """

    test_message = {
        "message_type": "Unknown",
    }

    on_data_called = False

    def on_data(data: aai.RealtimeTranscript):
        nonlocal on_data_called
        on_data_called = True

    on_error_called = False

    def on_error(error: aai.RealtimeError):
        nonlocal on_error_called
        on_error_called = True

    transcriber = aai.RealtimeTranscriber(
        on_data=on_data,
        on_error=on_error,
        sample_rate=44_100,
    )

    transcriber._impl._handle_message(test_message)

    assert not on_data_called
    assert not on_error_called


def test_create_temporary_token(httpx_mock: HTTPXMock):
    """
    Tests whether the creation of a temporary token is successful.
    """

    # mock the specific endpoint
    httpx_mock.add_response(
        url=f"{aai.settings.base_url}{ENDPOINT_REALTIME_TOKEN}",
        status_code=httpx.codes.OK,
        method="POST",
        json={"token": "123456"},
    )

    token = aai.RealtimeTranscriber.create_temporary_token(expires_in=3000)

    assert token == "123456"


# TODO: create tests for the `RealtimeTranscriber.close` method



from unittest.mock import mock_open, patch

import assemblyai as aai


def test_stream_file_empty_file():
    """
    Test streaming of an empty file.
    """

    data = b""
    sample_rate = 44100

    m = mock_open(read_data=data)

    with patch("builtins.open", m), patch("time.sleep", return_value=None):
        chunks = list(aai.extras.stream_file("fake_path", sample_rate))

    # Expect no chunk
    assert len(chunks) == 0


def test_stream_file_small_file():
    """
    Tests streaming a file smaller than 300ms.
    """

    data = b"\x00" * int(0.2 * 44100) * 2
    sample_rate = 44100

    m = mock_open(read_data=data)

    with patch("builtins.open", m), patch("time.sleep", return_value=None):
        chunks = list(aai.extras.stream_file("fake_path", sample_rate))

    # Expecting one chunks because of no padding at the end
    expected_chunk_length = int(0.2 * sample_rate * 2)
    assert len(chunks) == 1
    assert len(chunks[0]) == expected_chunk_length
    assert chunks[0] == b"\x00" * expected_chunk_length


def test_stream_file_large_file():
    """
    Test streaming a file larger than 300ms.
    """

    data = b"\x00" * int(0.6 * 44100) * 2
    sample_rate = 44100

    m = mock_open(read_data=data)

    with patch("builtins.open", m), patch("time.sleep", return_value=None):
        chunks = list(aai.extras.stream_file("fake_path", sample_rate))

    # Expecting two chunks
    assert len(chunks) == 2


def test_stream_file_exact_file():
    """
    Test streaming a file exactly 300ms long.
    """

    data = b"\x00" * int(0.3 * 44100) * 2
    sample_rate = 44100

    m = mock_open(read_data=data)

    with patch("builtins.open", m), patch("time.sleep", return_value=None):
        chunks = list(aai.extras.stream_file("fake_path", sample_rate))

    # Expecting one chunk
    assert len(chunks) == 1



import factory
import pytest
from pytest_httpx import HTTPXMock

import tests.unit.factories as factories
import tests.unit.unit_test_utils as test_utils
import assemblyai as aai

aai.settings.api_key = "test"


class SummarizationResponseFactory(factories.TranscriptCompletedResponseFactory):
    summary = factory.Faker("sentence")


@pytest.mark.parametrize("required_field", ["punctuate", "format_text"])
def test_summarization_fails_without_required_field(
    httpx_mock: HTTPXMock, required_field: str
):
    """
    Tests whether the SDK raises an error before making a request
    if `summarization` is enabled and the given required field is disabled
    """
    with pytest.raises(ValueError) as error:
        test_utils.submit_mock_transcription_request(
            httpx_mock,
            {},
            config=aai.TranscriptionConfig(
                summarization=True,
                **{required_field: False},  # type: ignore
            ),
        )

    # Check that the error message informs the user of the invalid parameter
    assert required_field in str(error)

    # Check that the error was raised before any requests were made
    assert len(httpx_mock.get_requests()) == 0


def test_summarization_disabled_by_default(httpx_mock: HTTPXMock):
    """
    Tests that excluding `summarization` from the `TranscriptionConfig` will
    result in the default behavior of it being excluded from the request body
    """
    mock_response = factories.generate_dict_factory(
        factories.TranscriptCompletedResponseFactory
    )()
    request_body, transcript = test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response,
        config=aai.TranscriptionConfig(),
    )

    # Check that request body was properly defined
    assert request_body.get("summarization") is None

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None
    assert transcript.summary is None


def test_default_summarization_params(httpx_mock: HTTPXMock):
    """
    Tests that including `summarization=True` in the `TranscriptionConfig`
    will result in `summarization=True` in the request body.
    """
    mock_response = factories.generate_dict_factory(SummarizationResponseFactory)()
    request_body, transcript = test_utils.submit_mock_transcription_request(
        httpx_mock, mock_response, aai.TranscriptionConfig(summarization=True)
    )

    # Check that request body was properly defined
    assert request_body.get("summarization") is True
    assert request_body.get("summary_model") is None
    assert request_body.get("summary_type") is None

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None
    assert transcript.summary == mock_response["summary"]


def test_summarization_with_params(httpx_mock: HTTPXMock):
    """
    Tests that including additional summarization parameters along with
    `summarization=True` in the `TranscriptionConfig` will result in all
    parameters being included in the request as well.
    """

    summary_model = aai.SummarizationModel.conversational
    summary_type = aai.SummarizationType.bullets

    mock_response = factories.generate_dict_factory(SummarizationResponseFactory)()

    request_body, transcript = test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response,
        aai.TranscriptionConfig(
            summarization=True,
            summary_model=summary_model,
            summary_type=summary_type,
        ),
    )

    # Check that request body was properly defined
    assert request_body.get("summarization") is True
    assert request_body.get("summary_model") == summary_model
    assert request_body.get("summary_type") == summary_type

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None
    assert transcript.summary == mock_response["summary"]


def test_summarization_params_excluded_when_disabled(httpx_mock: HTTPXMock):
    """
    Tests that additional summarization parameters are excluded from the submission
    request body if `summarization` itself is not enabled.
    """
    mock_response = factories.generate_dict_factory(
        factories.TranscriptCompletedResponseFactory
    )()
    request_body, transcript = test_utils.submit_mock_transcription_request(
        httpx_mock,
        mock_response,
        aai.TranscriptionConfig(
            summarization=False,
            summary_model=aai.SummarizationModel.conversational,
            summary_type=aai.SummarizationType.bullets,
        ),
    )

    # Check that request body was properly defined
    assert request_body.get("summarization") is None
    assert request_body.get("summary_model") is None
    assert request_body.get("summary_type") is None

    # Check that transcript was properly parsed from JSON response
    assert transcript.error is None
    assert transcript.summary is None



import time
from typing import BinaryIO, Generator, Optional
from warnings import warn

from . import api
from .client import Client


class AssemblyAIExtrasNotInstalledError(ImportError):
    def __init__(
        self,
        msg="""
        You must install the extras for this SDK to use this feature.
        Run `pip install "assemblyai[extras]"` to install the extras.
        Make sure to install `apt install portaudio19-dev` (Debian/Ubuntu) or
        `brew install portaudio` (MacOS) before installing the extras
        """,
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)


class MicrophoneStream:
    def __init__(self, sample_rate: int = 44_100, device_index: Optional[int] = None):
        """
        Creates a stream of audio from the microphone.

        Args:
            sample_rate: The sample rate to record audio at.
            device_index: The index of the input device to use. If None, uses the default device.
        """
        try:
            import pyaudio
        except ImportError:
            raise AssemblyAIExtrasNotInstalledError

        self._pyaudio = pyaudio.PyAudio()
        self.sample_rate = sample_rate

        self._chunk_size = int(self.sample_rate * 0.1)
        self._stream = self._pyaudio.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=sample_rate,
            input=True,
            frames_per_buffer=self._chunk_size,
            input_device_index=device_index,
        )

        self._open = True

    def __iter__(self):
        """
        Returns the iterator object.
        """

        return self

    def __next__(self):
        """
        Reads a chunk of audio from the microphone.
        """
        if not self._open:
            raise StopIteration

        try:
            return self._stream.read(self._chunk_size)
        except KeyboardInterrupt:
            raise StopIteration

    def close(self):
        """
        Closes the stream.
        """

        self._open = False

        if self._stream.is_active():
            self._stream.stop_stream()

        self._stream.close()
        self._pyaudio.terminate()


def stream_file(
    filepath: str,
    sample_rate: int,
) -> Generator[bytes, None, None]:
    """
    Mimics a stream of audio data by reading it chunk by chunk from a file.

    NOTE: Only supports WAV/PCM16 files as of now.

    Args:
        filepath: The path to the file to stream.
        sample_rate: The sample rate of the audio file.

    Returns: A generator that yields chunks of audio data.
    """
    chunk_duration = 0.3
    with open(filepath, "rb") as f:
        while True:
            # send in 300ms segments (2 bytes per frame)
            data = f.read(int(sample_rate * chunk_duration) * 2)

            if not data:
                break

            yield data

            time.sleep(chunk_duration)


def file_from_stream(data: BinaryIO) -> str:
    """
    DeprecationWarning: `file_from_stream()` is deprecated and will be removed in 1.0.0. Use `Transcriber.upload_file()` instead.

    Uploads the given stream and returns the uploaded audio url.

    This function can be used to transcribe data that's already
    available in memory.

    Example:
    ```
    upload_url = aai.extras.file_from_stream(data)

    transcriber = aai.Transcriber()
    transcript = transcriber.transcribe(upload_url)
    ```

    Args:
        `data`: A file-like object (in binary mode)
    """
    warn(
        "`file_from_stream()` is deprecated and will be removed in 1.0.0. Use `Transcriber.upload_file()` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return api.upload_file(
        client=Client.get_default().http_client,
        audio_file=data,
    )



import inspect

import pytest

import assemblyai as aai


def test_configuration_are_none_by_default():
    """
    Tests whether all configurations are None by default.
    """

    config = aai.TranscriptionConfig()
    fields = config.raw.__fields_set__ - {"language_code"}

    for name, value in inspect.getmembers(config):
        if name in fields and value is not None:
            pytest.fail(
                f"Configuration field {name} is {value} and not None by default."
            )
